import threading

import departure.board.board as board
import departure.board.movement as movement
import departure.board.departure_pb2 as departure_pb2
import departure.board.protobuf as protobuf


class BoardUpdater:
    def __init__(self, target_board: board.Board, board_lock: threading.RLock):
        self.board = target_board
        self.board_lock = board_lock

    def update(self):
        raise NotImplementedError()


class BoardUpdater_192_32_3_Rows(BoardUpdater):
    def __init__(self, target_board: board.Board, board_lock: threading.RLock):
        super().__init__(target_board, board_lock)

        self.board.sections = [
            board.BoardSection(0, 0, 192, 10, board.BoardSectionContent([], 0, 0)),
            board.BoardSection(0, 11, 192, 10, board.BoardSectionContent([], 0, 0)),
            board.BoardSection(0, 22, 192, 10, board.BoardSectionContent([], 0, 0)),
        ]

    def update_section(
        self,
        number: int,
        section_content: board.BoardSectionContent,
        section_movement: movement.Movement,
        section_continue_movement: bool,
    ):
        change_movement = True

        if section_content is not None:
            # if continuous movement is requested, do not change movement
            # if new content size is identical to old one
            if (
                section_continue_movement
                and section_content.content_w
                == self.board.sections[number]._content.content_w
                and section_content.content_h
                == self.board.sections[number]._content.content_h
            ):
                change_movement = False

            self.board.sections[number].update_content(section_content)

        if section_movement is not None and change_movement:
            self.board.sections[number].update_movement(section_movement)
            self.board.sections[number].reset()

    def update_sections(
        self,
        top_section_content: board.BoardSectionContent = None,
        top_section_movement: movement.Movement = None,
        top_section_continue_movement: bool = False,
        middle_section_content: board.BoardSectionContent = None,
        middle_section_movement: movement.Movement = None,
        middle_section_continue_movement: bool = False,
        bottom_section_content: board.BoardSectionContent = None,
        bottom_section_movement: movement.Movement = None,
        bottom_section_continue_movement: bool = False,
    ):

        with self.board_lock:
            self.update_section(
                0,
                top_section_content,
                top_section_movement,
                top_section_continue_movement,
            )

            self.update_section(
                1,
                middle_section_content,
                middle_section_movement,
                middle_section_continue_movement,
            )

            self.update_section(
                2,
                bottom_section_content,
                bottom_section_movement,
                bottom_section_continue_movement,
            )

    def update(self):
        raise NotImplementedError()


class BoardUpdater_192_32_3_Rows_Native(BoardUpdater_192_32_3_Rows):
    def update(
        self,
        top_section_content: board.BoardSectionContent = None,
        top_section_movement: movement.Movement = None,
        top_section_continue_movement: bool = False,
        middle_section_content: board.BoardSectionContent = None,
        middle_section_movement: movement.Movement = None,
        middle_section_continue_movement: bool = False,
        bottom_section_content: board.BoardSectionContent = None,
        bottom_section_movement: movement.Movement = None,
        bottom_section_continue_movement: bool = False,
    ):
        self.update_sections(
            top_section_content,
            top_section_movement,
            top_section_continue_movement,
            middle_section_content,
            middle_section_movement,
            middle_section_continue_movement,
            bottom_section_content,
            bottom_section_movement,
            bottom_section_continue_movement,
        )


class BoardUpdater_192_32_3_Rows_From_ProtocolBuffers(BoardUpdater_192_32_3_Rows):
    def update(
        self,
        board_sections_update_request: departure_pb2.BoardSectionsUpdateRequest,
    ) -> departure_pb2.BoardSectionsUpdateResponse:

        sections_update_data = {}
        sections_update_response = []

        for req in board_sections_update_request.requests:
            if req.section_index == 0:
                section_prefix = "top_section_"
            elif req.section_index == 1:
                section_prefix = "middle_section_"
            elif req.section_index == 2:
                section_prefix = "bottom_section_"
            else:  # invalid section
                sections_update_response.append(
                    # pylint: disable=E1101
                    departure_pb2.BoardSectionOperationStatus(
                        section_index=req.section_index,
                        status=departure_pb2.BoardSectionOperationStatus.SECTION_NOEXISTS,
                    )
                )
                continue

            # content
            sections_update_data[
                section_prefix + "content"
            ] = protobuf.deserialise_BoardSectionContent(req.content)

            # movement
            section_movement = protobuf.deserialise_RepeatedMovement(req.movement)
            if section_movement is not None:
                sections_update_data[section_prefix + "movement"] = section_movement

            # continue movement
            sections_update_data[
                section_prefix + "continue_movement"
            ] = req.continue_movement

            # response
            # pylint: disable=E1101
            sections_update_response.append(
                departure_pb2.BoardSectionOperationStatus(
                    section_index=req.section_index,
                    status=departure_pb2.BoardSectionOperationStatus.OK,
                )
            )

        self.update_sections(**sections_update_data)

        return departure_pb2.BoardSectionsUpdateResponse(
            status=sections_update_response
        )



import os

from zeep import Client

WSDL_FILENAME = "data/Wsiv.wsdl"


def soap_client():
    return Client(wsdl=os.path.join(os.path.dirname(__file__), WSDL_FILENAME))


def get_lines_realtime_realm():
    # may fail with requests.exceptions.ConnectionError
    return soap_client().service.getLines(line={"realm": "r"})


def get_lines_by_code(code_query):
    return soap_client().service.getLines(
        line={"code": f"*{code_query}*", "realm": "r"}
    )


def get_directions(line_id):
    return soap_client().service.getDirections(line={"id": line_id})


def get_stations_by_line(line_id):
    return soap_client().service.getStations(station={"line": {"id": line_id}}).stations


def get_stations_by_name(name_query):
    return soap_client().service.getStations(
        station={"name": name_query, "line": {"realm": "r"}}
    )


def get_missions_next(line_id, station_id, direction_sens):
    return soap_client().service.getMissionsNext(
        station={"line": {"id": line_id}, "id": station_id},
        direction={"sens": direction_sens},
    )



import pathlib
import csv


STATION_CODES_FILENAME = "data/station_codes.csv"


def _stations(station_codes_filename):
    station_codes_path = pathlib.Path(__file__).parents[0] / station_codes_filename

    # read from stations data file if available
    with station_codes_path.open() as stations_codes_file:
        reader = csv.reader(stations_codes_file)
        next(reader, None)  # skip header row
        stations = {}
        for row in reader:
            stations[row[1]] = row[0]

    return stations


STATIONS = _stations(STATION_CODES_FILENAME)



from tabulate import tabulate


def list_lines(lines):
    table = []
    for line in sorted(lines, key=lambda k: k["code"]):
        table.append([line["id"], f"{line['reseau']} {line['code']}", line["name"]])

    print(tabulate(table, headers=["id", "code", "nom"]))


def list_directions(directions):
    print(
        tabulate(
            [[direction, directions[direction]] for direction in directions.keys()],
            headers=["code", "direction"],
        )
    )


def list_stations(stations, show_line: bool = False, show_station_id: bool = False):
    table = []

    # populate table
    for station in sorted(stations, key=lambda k: k["name"]):
        station_item = []

        if show_line:
            station_item.append(station["line"]["id"])
            station_item.append(
                f"{station['line']['reseau']} {station['line']['code']}"
            )

        if show_station_id:
            station_item.append(station["line_station_id"])

        station_item.append(station["name"])

        table.append(station_item)

    # set table headers
    headers = []
    if show_line:
        headers.extend(["line id", "line code"])
    if show_station_id:
        headers.append("line station id")
    headers.append("station name")

    # display table
    print(tabulate(table, headers))


def list_departures(departures):
    if departures["missions"]:
        print(tabulate(departures["missions"], tablefmt="plain"))
    if departures["perturbations"]:
        print(f"! {departures['perturbations']}")



import logging
from typing import List, Any
import re

from . import data, api, commons

logger = logging.getLogger(__name__)


def check_params(line_ids: List[str] = None, station_id=None, direction=None):
    if not any((line_ids, station_id, direction)):
        return

    stations = data.STATIONS

    if station_id is not None:
        try:
            station = stations[station_id]
        except Exception as e:
            raise commons.TflTubeException(f"invalid station id {station_id}") from e

    if line_ids is not None:
        for line_id in line_ids:
            if line_id not in data.lines:
                raise commons.TflTubeException(f"invalid line id {line_id}")
            if (station_id is not None) and (line_id not in station["lines"]):
                raise commons.TflTubeException(
                    f"station/line mismatch for {station_id}/{line_id}"
                )

    if direction is not None:
        for line_id in line_ids:
            if direction not in data.lines[line_id]["canonical_directions"]:
                raise commons.TflTubeException(
                    f"invalid direction {direction} for {line_id}"
                )


def stations_by_string(string: str) -> dict:
    string = str(string).lower()
    results = {}
    stations = data.STATIONS

    # iterate over stations
    for station_id in stations:
        # match?
        if string in stations[station_id]["name"].lower():
            results[station_id] = stations[station_id]

    return results


def lines() -> list:
    return {line_id: data.lines[line_id]["name"] for line_id in data.lines}


def stations_by_line(line_id: str) -> dict:
    results = {}
    stations = data.STATIONS
    stoppoints = data.stoppoints_by_line(line_id)

    for stoppoint in stoppoints:
        results[stoppoint["naptanId"]] = stations[stoppoint["naptanId"]]

    return results


def directions(line_id: str) -> dict:
    return list(data.lines[line_id]["canonical_directions"].keys())


def location_platform_from_current_location(current_location: str):
    """Parses currentLocation: "At ..." -> (location, platform_number)"""

    # ignore if not *at* a location
    if not current_location.startswith("At "):
        return None, None

    # discard initial 'At '
    current_location = current_location[3:]

    # at current station/platform (i.e. "[At ]Platform")
    if current_location == "Platform":
        return "<current>", "<current>"  # special values

    # with station and platform, supporting the following alternatives
    # - "Manor House Platform 1" (normal)
    # - "East Finchley, Platform 1" (extra comma)
    # - "London Bridge Plaform 1" (typo)
    # - "Kings Cross P7" (abbreviation)
    station_platform = re.match(r"^(.*?),? P(?:lat?form )?(\d+)$", current_location)

    if station_platform is not None:
        return station_platform.group(1), station_platform.group(2)

    # without platform number (e.g. "[At ]Angel platform")
    if current_location.endswith(" platform"):
        return current_location[: -len(" platform")], None

    # no mention of platform
    if "Platform" not in current_location:
        return current_location, None

    return None, None


def station_from_location(location_name: str, current_station_id: str, line_id: str):
    if location_name is None:
        return None

    # actual stations
    if location_name in data.STATION_ID_BY_NAME:
        return data.STATION_ID_BY_NAME[location_name]

    if location_name == "<current>":
        return current_station_id

    # alternative stations names
    if location_name in data.alternative_station_names:
        return data.alternative_station_names[location_name]

    # ambiguous alternative stations names (resolve through line)
    if location_name in data.alternative_station_names_ambiguous:
        return data.alternative_station_names_ambiguous[location_name][line_id]

    # known non-stations
    if location_name in data.non_station_locations:
        return None

    logger.warning("unresolved location %s", location_name)
    return None


def filter_trains_by_direction_of_departure(
    arrivals: List[Any],
    requested_direction: str,
    conservative: bool = False,  # also match uncertain directions
):
    trains = {}

    for arrival in arrivals:
        # extract train id
        line_id = arrival["lineId"]
        vehicle_id = arrival["vehicleId"]

        # ignore duplicate trains
        train_id = f"{line_id}-{vehicle_id}"
        if train_id in trains:
            continue

        # extract train data
        station_id = arrival["naptanId"]
        towards_station_id = station_id_from_towards(arrival["towards"], line_id)
        platform_departure_direction = departure_direction_from_platform_name(
            arrival["platformName"]
        )
        train_canonical_direction = arrival.get("direction", None)
        destination_station_id = arrival.get("destinationNaptanId", None)

        # check if will depart in requested direction
        will_depart_in_requested_direction = (
            will_arriving_train_depart_from_station_in_direction(
                line_id,
                station_id,
                requested_direction,
                towards_station_id,
                platform_departure_direction=platform_departure_direction,
                train_canonical_direction=train_canonical_direction,
                destination_station_id=destination_station_id,
            )
        )

        # ignore if not departing in requested direction
        if not will_depart_in_requested_direction:
            continue

        # ignore if can't determine departing direction and conservative
        if will_depart_in_requested_direction is None and conservative:
            continue

        train = {
            "line_id": line_id,
            "vehicle_id": vehicle_id,
            "station_id": station_id,
            "towards": arrival["towards"],
            "towards_station_id": towards_station_id,
            "platform_name": arrival["platformName"],
            "platform_number": number_from_platform_name(arrival["platformName"]),
            "destination_station_id": destination_station_id,
            "time_to_station": arrival["timeToStation"],
        }

        trains[train_id] = train

    return trains


def current_station_platform_number_from_current_location(
    current_location: str, line_id: str, station_id: str, platform_name: str
):
    at_location, at_platform_number = location_platform_from_current_location(
        current_location
    )

    if at_platform_number == "<current>":
        at_platform_number = number_from_platform_name(platform_name)

    at_station_id = station_from_location(at_location, station_id, line_id)

    return at_station_id, at_platform_number


def filter_trains_currently_at_platform(
    arrivals: List[Any],
    line_ids: List[str] = None,
    requested_station_id: str = None,
    requested_direction: str = None,
):
    trains = {}
    processed_trains = set()

    for arrival in arrivals:
        # extract train id
        line_id = arrival["lineId"]
        vehicle_id = arrival["vehicleId"]

        # ignore duplicate trains
        train_id = f"{line_id}-{vehicle_id}"
        if train_id in processed_trains:
            continue
        else:
            processed_trains.add(train_id)

        # extract train data
        line_id = arrival["lineId"]
        station_id = arrival["naptanId"]
        platform_departure_direction = departure_direction_from_platform_name(
            arrival["platformName"]
        )
        train_canonical_direction = arrival.get("direction", None)
        destination_station_id = arrival.get("destinationNaptanId", None)
        towards_station_id = station_id_from_towards(arrival["towards"], line_id)

        # get current station id
        (
            current_station_id,
            current_platform,
        ) = current_station_platform_number_from_current_location(
            arrival["currentLocation"], line_id, station_id, arrival["platformName"]
        )

        # ignore if train not at a station
        if current_station_id is None:
            continue

        # ignore if train is not on one of the requested lines
        if line_ids is not None and line_id not in line_ids:
            continue

        # ignore if train is not at requested station
        if (
            requested_station_id is not None
            and current_station_id != requested_station_id
        ):
            continue

        # ignore if train not in requested direction
        if (
            requested_direction is not None
            and not will_arriving_train_depart_from_current_location_in_direction(
                current_station_id,
                line_id,
                station_id,
                requested_direction,
                towards_station_id,
                platform_departure_direction,
                train_canonical_direction,
                destination_station_id,
            )
        ):
            continue

        # create item for train at platform
        train = {
            "line_id": line_id,
            "vehicle_id": vehicle_id,
            "station_id": current_station_id,
            "towards": arrival["towards"],
            "towards_station_id": towards_station_id,
            "platform_number": current_platform,
            "destination_station_id": destination_station_id,
            "time_to_station": 0,
        }

        trains[train_id] = train

    return trains


def next_trains(line_id: str, station_id: str, requested_direction: str = None):
    # check parameters
    check_params([line_id], station_id, requested_direction)

    # include merged lines
    line_ids = [line_id] + data.lines[line_id].get("merged_lines", [])

    # get trains arriving at station
    arrivals = api.line_arrivals(line_ids, station_id=station_id)
    if arrivals is None:
        return {}

    # filter by direction if requested
    if requested_direction is not None:
        trains = filter_trains_by_direction_of_departure(arrivals, requested_direction)

    # get all trains including those at station of interest
    arrivals = api.tube_arrivals(count=2)  # use -1 for debugging
    if arrivals is None:
        return trains

    # filter by current station (and direction of departure if requested)
    trains_currently_at_platform = filter_trains_currently_at_platform(
        arrivals, line_ids, station_id, requested_direction
    )

    # merge next and current (will overwrite next)
    trains.update(trains_currently_at_platform)

    return trains


def number_from_platform_name(platform_name: str):
    if "Platform " in platform_name:
        return platform_name.split("Platform ")[1]
    return None


def departure_direction_from_platform_name(platform_name: str):
    # platform name contains both direction and platform number
    for prefix in data.platform_name_prefix_direction:
        if platform_name.startswith(prefix):
            return data.platform_name_prefix_direction[prefix]

    # otherwise need to figure out things differently
    return None


def canonical_direction_from_platform_direction(line_id, direction):
    if line_id not in data.lines or line_id == "circle":
        return None

    try:
        return data.lines[line_id]["canonical_directions"][direction]
    except Exception:  # pylint: disable=broad-except
        pass  # passthrough to default

    return None


def circle_line_rail_arriving_at_station(
    station_id: str,
    platform_departure_direction: str = None,
    destination_station_id: str = None,
    towards_station_id: str = None,
):
    # if destination is known
    if destination_station_id is not None:
        try:
            return data.lines["circle"]["termini"][destination_station_id][
                "arrival_rail"
            ]
        except Exception:  # pylint: disable=broad-except
            pass

    # if towards is known
    if towards_station_id is not None:
        try:
            return data.lines["circle"]["termini"][towards_station_id]["arrival_rail"]
        except Exception:  # pylint: disable=broad-except
            pass

    # if direction is known from platform name (should cover remaining cases)
    if platform_departure_direction == "Westbound":
        if station_id in data.circle_line_outer_rail_westbound_stations:
            return "Outer Rail"
        else:
            return "Inner Rail"
    elif platform_departure_direction == "Eastbound":
        if station_id in data.circle_line_outer_rail_westbound_stations:
            return "Inner Rail"
        else:
            return "Outer Rail"

    # should never happen
    return None


def inverse_canonical_direction(canonical_direction):
    return {"inbound": "outbound", "outbound": "inbound"}.get(canonical_direction, None)


def canonical_direction_of_train_from_station_terminating_at(
    line_id, station_id, terminus_station_id
):
    # normal terminating direction
    try:
        terminating_canonical_direction = data.lines[line_id]["canonical_directions"][
            data.lines[line_id]["termini"][terminus_station_id]["arrival"]
        ]
    except Exception:  # pylint: disable=broad-except
        return None  # not a terminus

    # handle stations with inverse terminating directions at inline stations,
    # e.g. trains from EBY/central at end of service, terminating at WYC
    # (normally a Westbound inline terminus)
    try:
        if (
            data.lines[line_id]["termini"][terminus_station_id]["is_inline"]
            and station_id
            in data.lines[line_id]["termini"][terminus_station_id][
                "stations_inverse_terminating_direction"
            ]
        ):
            terminating_canonical_direction = inverse_canonical_direction(
                terminating_canonical_direction
            )
    except Exception:  # pylint: disable=broad-except
        pass

    return terminating_canonical_direction


def resolve_train_canonical_direction(
    line_id: str,
    station_id: str,
    towards_station_id: str = None,
    destination_station_id: str = None,
    platform_direction: str = None,
    train_canonical_direction: str = None,
):
    """
    not at terminus: arrival == departure
    """

    current_canonical_direction = None
    are_canonical_directions_coherent = True

    # resolve from towards
    if towards_station_id is not None:
        towards_canonical_direction = (
            canonical_direction_of_train_from_station_terminating_at(
                line_id, station_id, towards_station_id
            )
        )
        if towards_canonical_direction is not None:
            # update resolved direction
            current_canonical_direction = towards_canonical_direction

    # resolve from destination
    if destination_station_id is not None:
        destination_canonical_direction = (
            canonical_direction_of_train_from_station_terminating_at(
                line_id, station_id, destination_station_id
            )
        )

        if destination_canonical_direction is not None:
            # update resolved direction and coherence check
            if current_canonical_direction is None:
                current_canonical_direction = destination_canonical_direction
            else:
                are_canonical_directions_coherent &= (
                    current_canonical_direction == destination_canonical_direction
                )

    # resolve from platform
    if platform_direction is not None:
        platform_canonical_direction = canonical_direction_from_platform_direction(
            line_id, platform_direction
        )

        # update resolved direction and coherence check
        if current_canonical_direction is None:
            current_canonical_direction = platform_canonical_direction
        else:
            are_canonical_directions_coherent &= (
                current_canonical_direction == platform_canonical_direction
            )

    # use declared canonical direction otherwise
    # e.g. at LGN central inbound 'Platform 2/3'
    #      or WCY central outbound 'Platform 2/3'
    if train_canonical_direction is not None:
        # inverse declared canonical direction if north of HLT on HLT loop
        # (see data)
        if station_id in data.hainault_loop_outer_rail_outbound_stations:
            train_canonical_direction = inverse_canonical_direction(
                train_canonical_direction
            )

        if current_canonical_direction is None:
            current_canonical_direction = train_canonical_direction
        else:
            are_canonical_directions_coherent &= (
                current_canonical_direction == train_canonical_direction
            )

    if not are_canonical_directions_coherent:
        logger.warning(
            "incoherent direction at %s (%s): "
            "towards %s / destination %s / %s platform / %s",
            station_id,
            line_id,
            destination_station_id,
            towards_station_id,
            platform_direction,
            train_canonical_direction,
        )

    return current_canonical_direction


def will_arriving_train_depart_from_station_in_direction(
    line_id: str,
    station_id: str,
    requested_direction: str,
    towards_station_id: str,
    platform_departure_direction: str = None,
    train_canonical_direction: str = None,
    destination_station_id: str = None,
):
    # 1) check if current station is terminus
    is_station_terminus = is_station_terminus_for_train(
        line_id, station_id, destination_station_id, towards_station_id
    )

    # general case: won't depart if current station is terminus
    if is_station_terminus:
        return False

    # special case: if station *could* be a terminus but can't tell due to
    # missing information, consider that the station isn't a terminus
    if is_station_terminus is None:
        pass  # replace with return None (can't tell) or False (no) as needed

    # 2) on Circle Line (which only has one canonical direction for
    # all directions), resolve via inner/outer rail
    if line_id == "circle":
        # departure direction == arrival direction as not a terminus
        station_rail = circle_line_rail_arriving_at_station(
            station_id,
            platform_departure_direction=platform_departure_direction,
            destination_station_id=destination_station_id,
            towards_station_id=towards_station_id,
        )
        # can't determine
        if station_rail is None:
            return None

        requested_rail = circle_line_rail_arriving_at_station(
            station_id, platform_departure_direction=requested_direction
        )

        return station_rail == requested_rail

    # 3) general case: resolve via canonical direction
    requested_canonical_direction = canonical_direction_from_platform_direction(
        line_id, requested_direction
    )

    # get actual train canonical direction
    actual_train_canonical_direction = resolve_train_canonical_direction(
        line_id,
        station_id,
        towards_station_id=towards_station_id,
        destination_station_id=destination_station_id,
        platform_direction=platform_departure_direction,
        train_canonical_direction=train_canonical_direction,
    )

    if actual_train_canonical_direction is not None:
        return requested_canonical_direction == actual_train_canonical_direction

    # can't determine direction
    return None


def will_arriving_train_depart_from_current_location_in_direction(
    current_station_id: str,
    line_id: str,
    station_id: str,
    requested_direction: str,
    towards_station_id: str,
    platform_departure_direction: str = None,
    train_canonical_direction: str = None,
    destination_station_id: str = None,
):
    # 1) check if current station is terminus in arriving direction
    # test cases:
    # - depart from EBY towards HLT ~> will depart
    # - depart from EBY towards WCY ~> should depart (WTT id 110 p. 125)
    #   >> incoherent direction at 940GZZLUEBY (central):
    #      destination 940GZZLUWCY / towards 940GZZLUWCY / Eastbound platform
    #      / outbound
    is_current_station_terminus = is_station_terminus_for_train(
        line_id, current_station_id, destination_station_id, towards_station_id
    )

    # general case: won't depart if current station is terminus
    if is_current_station_terminus:
        return False

    # special case: if current station *could* be a terminus but can't tell
    # due to missing information, consider that the current station isn't a
    # terminus
    if is_current_station_terminus is None:
        pass  # replace with return None (can't tell) or False (no) as needed

    # 2) determine platform arrival direction
    is_station_terminus = is_station_terminus_for_train(
        line_id, station_id, destination_station_id, towards_station_id
    )

    # if train terminates at station, get arrival direction from terminus data
    if is_station_terminus:
        try:
            platform_arrival_direction = data.lines[line_id]["termini"][station_id][
                "arrival"
            ]
        except Exception:  # pylint: disable=broad-except
            return False
    # otherwise arrival direction and departure direction are identical
    else:
        platform_arrival_direction = platform_departure_direction

    # special case: if station *could* be a terminus but can't tell due to
    # missing information, consider that the station isn't a terminus
    if is_station_terminus is None:
        pass  # replace with return None (can't tell) or add as and to
        # previous if statement (consider as terminus)

    # 3) on Circle Line (which only has one canonical direction for all
    # directions), resolve directions via inner/outer rail
    if line_id == "circle":
        # departure direction == arrival direction as not a terminus
        station_rail = circle_line_rail_arriving_at_station(
            station_id,
            platform_departure_direction=platform_arrival_direction,
            destination_station_id=destination_station_id,
            towards_station_id=towards_station_id,
        )
        # can't determine
        if station_rail is None:
            return None

        requested_rail = circle_line_rail_arriving_at_station(
            current_station_id,
            platform_departure_direction=requested_direction,
        )
        return station_rail == requested_rail

    # 4) general case: resolve via canonical direction
    requested_canonical_direction = canonical_direction_from_platform_direction(
        line_id, requested_direction
    )

    # get actual train canonical direction
    actual_train_canonical_direction = resolve_train_canonical_direction(
        line_id,
        station_id,
        towards_station_id=towards_station_id,
        destination_station_id=destination_station_id,
        platform_direction=platform_arrival_direction,
        train_canonical_direction=train_canonical_direction,
    )

    if actual_train_canonical_direction is not None:
        return requested_canonical_direction == actual_train_canonical_direction

    # 5) can't determine direction
    return None


def station_id_from_station_name(station_name: str, line_id: str):
    try:
        return data.STATION_ID_BY_NAME[station_name]
    except Exception:  # pylint: disable=broad-except
        pass

    try:
        return data.alternative_station_names[station_name]
    except Exception:  # pylint: disable=broad-except
        pass

    try:
        return data.alternative_station_names_ambiguous[station_name][line_id]
    except Exception:  # pylint: disable=broad-except
        pass

    return None


def station_id_from_towards(towards: str, line_id: str):
    if towards == "Check Front of Train":
        return None

    try:
        return data.alternative_terminus_names[towards]
    except Exception:  # pylint: disable=broad-except
        pass

    if " via " in towards:
        station_name = towards.split(" via ")[0]
    elif " Via " in towards:
        station_name = towards.split(" Via ")[0]
    else:
        station_name = towards

    return station_id_from_station_name(station_name, line_id)


def is_station_terminus_for_train(
    line_id,
    station_id,
    destination_station_id: str = None,
    towards_station_id: str = None,
):
    # if destination is known and same as station
    if destination_station_id is not None:
        return station_id == destination_station_id

    # if towards (e.g. "Edgware via CX") resolves to requested station
    if towards_station_id is not None:
        return station_id == towards_station_id

    # if station is a terminus
    if station_id in data.lines[line_id]["termini"]:
        # can't tell if inline (e.g. BKG/HLT), no destinationNaptanId,
        # towards="Check Front of Train"
        if data.lines[line_id]["termini"][station_id]["is_inline"]:
            return None
        # non-inline (e.g. WIM)
        return True

    # default: not a terminus (caveat: false negative if abnormal service
    # with no information on destination)
    return False



import json
import os
import logging

import requests

from . import commons

logger = logging.getLogger(__name__)


def api_request(url: str):
    commons.check_env_vars()

    try:
        response = requests.get(
            url,
            timeout=15,
            # http://doc.navitia.io/#authentication
            auth=(os.environ["SNCF_KEY"], ""),
        )
    except requests.exceptions.Timeout:
        logger.warning("SNCF API HTTP request timed out")
        return None
    except Exception as e:  # pylint: disable=broad-except
        logger.warning("SNCF API HTTP request error: %s", str(e))
        return None

    # process response
    if response.status_code == 200:
        return json.loads(response.content.decode("utf-8"))

    return None


def departures(station_id):
    url = (
        "https://api.navitia.io/v1/coverage/sncf/stop_areas/"
        f"stop_area:OCE:SA:{station_id}/departures"
    )
    return api_request(url)


def disruptions(disruption_id):
    url = f"https://api.navitia.io/v1/coverage/sncf/disruptions/{disruption_id}/"
    return api_request(url)


def vehicle_journeys(vehicle_journey_id):
    url = f"https://api.navitia.io/v1/coverage/sncf/vehicle_journeys/{vehicle_journey_id}/"
    return api_request(url)



import departure.board.board as board
import departure.board.movement as movement

# pylint: disable=attribute-defined-outside-init,protected-access


class TestContentUpdateWithMovement:
    def setup_method(self):
        self.content_pixels = [
            [0, 0, (255, 0, 0)],
            [0, 7, (0, 255, 0)],
            [0, 15, (0, 0, 255)],
        ]

        self.next_content_pixels = [
            [0, 0, (255, 0, 0)],
            [1, 0, (255, 0, 0)],
            [0, 7, (0, 255, 0)],
            [1, 7, (0, 255, 0)],
            [0, 15, (0, 0, 255)],
            [1, 15, (0, 0, 255)],
        ]

        self.board = board.Board(
            board_w=8,
            board_h=8,
            sections=[
                board.BoardSection(
                    output_x=0,
                    output_y=0,
                    output_w=8,
                    output_h=8,
                    content=board.BoardSectionContent(
                        content_pixels=self.content_pixels,
                        content_w=16,
                        content_h=16,
                    ),
                    movement=movement.ScrollingContent(30, 0, 0, -1),
                )
            ],
        )

        self.next_content = board.BoardSectionContent(
            content_pixels=self.next_content_pixels, content_w=16, content_h=16
        )

    def test_init(self):
        self.board.reset()  # includes first move

        # has scrolled up by 1px
        assert self.board.pixels() == [
            [0, 6, (0, 255, 0)],
        ]

    def test_30_ms(self):
        self.board.reset()
        self.board.update(30)

        # has scrolled up by 2px
        assert self.board.pixels() == [
            [0, 5, (0, 255, 0)],
        ]

    def test_30_ms_update_content(self):
        self.board.reset()
        self.board.update(30)

        self.board.sections[0].update_content(self.next_content)
        assert self.board.pixels() == [
            [0, 5, (0, 255, 0)],
            [1, 5, (0, 255, 0)],
        ]


class TestBoardSection:
    def test_update_movement(self):
        movement1 = movement.NoMovement()
        movement2 = movement.StaticContent(0)
        board_section = board.BoardSection(
            0,
            0,
            8,
            8,
            board.BoardSectionContent([(0, 0, (255, 255, 255))], 8, 8),
            movement=movement1,
        )

        board_section.reset()
        _ = board_section.pixels()

        assert board_section._movement == movement1
        assert not board_section._has_changed

        board_section.update_movement(movement2)

        assert board_section._movement == movement2
        assert board_section._has_changed



import click

from departure.commons.log import init_logging
from departure.provider.ns import cli as ns
from departure.provider.national_rail import cli as national_rail
from departure.provider.ratp import cli as ratp
from departure.provider.sncf import cli as sncf
from departure.provider.tfl_tube import cli as tfl_tube
from departure.provider.transilien import cli as transilien


# initialise logging
init_logging()


# CLI entry point
@click.group()
def entry_point():
    """
    Get station information and departures from public transport operators, and update
    a departure board with departure information for a station.
    """


# add commands
entry_point.add_command(tfl_tube.cli, name="lu")
entry_point.add_command(ns.cli, name="ns")
entry_point.add_command(national_rail.cli, name="national-rail")
entry_point.add_command(ratp.cli, name="ratp")
entry_point.add_command(sncf.cli, name="sncf")
entry_point.add_command(transilien.cli, name="transilien")


if __name__ == "__main__":
    entry_point()



import os

import uvicorn
import click


@click.command()
def start():
    uvicorn_reload = "UVICORN_RELOAD" in os.environ
    uvicorn.run("departure.web.api_server:app", reload=uvicorn_reload, host="0.0.0.0")


if __name__ == "__main__":
    start()



import logging  # for type hinting
import io
from contextlib import redirect_stdout
from typing import Callable, Any


def log_function_stdout_to_debug(
    logger: logging.Logger, function: Callable[..., None], *argv: Any
):
    stdout_string = io.StringIO()
    with redirect_stdout(stdout_string):
        function(*argv)
    logger.debug(stdout_string.getvalue())


def init_logging(logger_level=logging.DEBUG, handler_level=logging.DEBUG):
    logger = logging.getLogger("departure")
    logger.setLevel(level=logger_level)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(handler_level)

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    console_handler.setFormatter(formatter)

    logger.addHandler(console_handler)

    return logger



import time
import threading
import logging

import departure.board.data_updater as data_updater
from departure.commons.log import log_function_stdout_to_debug
from . import national_rail, view_model, ui

logger = logging.getLogger(__name__)


class DataUpdaterNationalRail(data_updater.DataUpdater):
    def __init__(
        self,
        target_view_model: view_model.ViewModelNationalRail,
        data_refresh_delay_in_s: int,
        end_event: threading.Event,
        station_id: str,
    ):
        super().__init__(target_view_model, data_refresh_delay_in_s, end_event)
        self.station_id = station_id
        self.services = None

    def update(self):
        update_start_time = time.monotonic()
        logger.info("data refresh started at %s", update_start_time)

        try:
            services = national_rail.next_services(self.station_id)

            # debugging: capture list of services to string, then log
            log_function_stdout_to_debug(logger, ui.list_services, services)

            # update board only if there were changes
            if services == self.services:
                logger.info("no change - no update sent to board")
            else:
                self.view_model.update(services)
                self.services = services

        except ConnectionError:
            logger.warning(
                "connection to National Rail failed after %s",
                time.monotonic() - update_start_time,
            )

        logger.info("data refresh duration %s", time.monotonic() - update_start_time)



from tabulate import tabulate

from . import data


def list_stations(stations=None):
    # default: list all stations
    if stations is None:
        stations = data.STATIONS

    print(
        tabulate(
            [
                [
                    station_id,
                    stations[station_id]["nom"],
                    stations[station_id]["libelle"],
                ]
                for station_id in sorted(stations)
            ],
            headers=["id", "name", "label"],
        )
    )

    print(f"> total: {len(stations)}")


def list_trains(trains):
    print(
        tabulate(
            [
                [train["mission"], train["time"], train["terminus"]]
                for train in trains
            ],
            tablefmt="plain"
        )
    )



from typing import List


class MovementException(Exception):
    pass


class Movement:
    def __init__(self, x_offset_init=0, y_offset_init=0):
        self.x_offset_init = x_offset_init
        self.y_offset_init = y_offset_init
        self.x_offset = None
        self.y_offset = None
        self._started = False

    def reset(self):
        # next line removed to avoid complications with inheritance
        # self.set_offset(x_offset, y_offset)
        self.x_offset = self.x_offset_init
        self.y_offset = self.y_offset_init
        self._started = True

    def update(self, delta_time):
        # raise Exception if update() is called before reset()
        if not self._started:
            raise MovementException(f"movement not started for object {type(self)}")
        return False  # have offsets changed?

    def set_offset(self, x_offset, y_offset):
        # e.g. used to update movement offset if it has wrapped
        self.x_offset = x_offset
        self.y_offset = y_offset


class NoMovement(Movement):
    pass


class MovementSegment(Movement):
    # abstract parent class for Movement segments (i.e. Movements that
    # can have a limited duration, designed to be combined with other
    # Movements)
    def __init__(self, x_offset_init=0, y_offset_init=0):
        super().__init__(x_offset_init, y_offset_init)
        self.elapsed_time = None
        self.extra_time_after_end = None
        self.end_of_segment_reached = None

    def reset(self):
        super().reset()
        # used at beginning of segment and after cycling back to segment
        self.elapsed_time = 0
        self.extra_time_after_end = 0
        self.end_of_segment_reached = False

    def update(self, delta_time):
        return super().update(delta_time)


class StaticContent(MovementSegment):
    def __init__(self, total_duration, x_offset_init=0, y_offset_init=0):
        super().__init__(x_offset_init, y_offset_init)
        self.total_duration = total_duration  # 0: forever

    def update(self, delta_time):
        super().update(delta_time)

        # no update if unlimited duration
        if self.total_duration == 0:
            return False

        self.elapsed_time += delta_time

        # if end of segment reached, also calculate remaining time beyond end
        if self.elapsed_time >= self.total_duration:
            self.end_of_segment_reached = True
            self.extra_time_after_end = self.elapsed_time - self.total_duration
            self.elapsed_time = self.total_duration

        return False


class ScrollingContent(MovementSegment):
    def __init__(
        self,
        step_duration: int,
        total_steps: int,  # 0: forever
        delta_x_per_step: int,
        delta_y_per_step: int,
        x_offset_init: int = 0,
        y_offset_init: int = 0,
    ):
        super().__init__(x_offset_init, y_offset_init)
        self.step_duration = step_duration
        self.total_steps = total_steps
        self.total_duration = total_steps * step_duration  # helper
        self.delta_x_per_step = delta_x_per_step
        self.delta_y_per_step = delta_y_per_step
        self.current_step_num = None

    def reset(self):
        super().reset()
        self.current_step_num = 0

        # first scroll at t==0 (will not scroll at t==total_duration)
        self.x_offset += self.delta_x_per_step
        self.y_offset += self.delta_y_per_step

    def update(self, delta_time):
        super().update(delta_time)
        new_elapsed_time = self.elapsed_time + delta_time

        # if perpetual scroll
        if self.total_duration == 0:
            # advance...
            new_step_num = new_elapsed_time // self.step_duration
            delta_step_num = new_step_num - self.current_step_num
            self.x_offset += delta_step_num * self.delta_x_per_step
            self.y_offset += delta_step_num * self.delta_y_per_step

            # ... and reset to step 0 (avoids risk of overflow)
            self.current_step_num = 0
            self.elapsed_time = new_elapsed_time % self.step_duration
            return delta_step_num > 0

        # if end of segment reached
        # note - '>=': do *not* scroll at t==total_duration (scrolled at t==0)
        if new_elapsed_time >= self.total_duration:
            self.end_of_segment_reached = True

            # advance
            delta_step_num = (self.total_steps - 1) - self.current_step_num
            self.x_offset += delta_step_num * self.delta_x_per_step
            self.y_offset += delta_step_num * self.delta_y_per_step

            # set current position at end of segment
            self.current_step_num = self.total_steps - 1
            self.elapsed_time = self.total_duration

            # and calculate remaining time beyond end
            self.extra_time_after_end = new_elapsed_time - self.total_duration

            return delta_step_num > 0

        # if within (bounded) segment
        new_step_num = new_elapsed_time // self.step_duration
        delta_step_num = new_step_num - self.current_step_num
        self.x_offset += delta_step_num * self.delta_x_per_step
        self.y_offset += delta_step_num * self.delta_y_per_step
        self.current_step_num = new_step_num
        self.elapsed_time = new_elapsed_time

        return delta_step_num > 0


class MovementCycle(Movement):
    def __init__(self, anim_sequence: List[MovementSegment]):
        super().__init__()
        self.anim_sequence = anim_sequence
        self.cur_segment = None
        self._started = False

    def reset(self):
        super().reset()
        self.cur_segment = 0
        self.anim_sequence[0].reset()

        # get initial offset (content may move after reset)
        self.x_offset = self.anim_sequence[0].x_offset
        self.y_offset = self.anim_sequence[0].y_offset

        self._started = True

    def update(self, delta_time):
        super().update(delta_time)

        while True:
            # update current segment by delta_time
            self.anim_sequence[self.cur_segment].update(delta_time)
            end_of_segment_reached = self.anim_sequence[
                self.cur_segment
            ].end_of_segment_reached
            new_x_offset = self.anim_sequence[self.cur_segment].x_offset
            new_y_offset = self.anim_sequence[self.cur_segment].y_offset

            # stop if end of current segment not reached
            if not end_of_segment_reached:
                break

            # otherwise retrieve amount of remaining time...
            delta_time = self.anim_sequence[self.cur_segment].extra_time_after_end

            # ... and move on to next segment
            self.cur_segment = (self.cur_segment + 1) % len(self.anim_sequence)
            self.anim_sequence[self.cur_segment].x_offset_init = new_x_offset
            self.anim_sequence[self.cur_segment].y_offset_init = new_y_offset
            self.anim_sequence[self.cur_segment].reset()

        if new_x_offset == self.x_offset and new_y_offset == self.y_offset:
            return False

        self.set_offset(new_x_offset, new_y_offset)
        return True

    def set_offset(self, x_offset, y_offset):
        super().set_offset(x_offset, y_offset)
        self.anim_sequence[self.cur_segment].set_offset(x_offset, y_offset)



import pytest

import departure.board.movement as movement

# pylint: disable=protected-access


class TestStaticContent:
    def test_init(self):
        static_content = movement.StaticContent(50)
        assert static_content.x_offset is None
        assert static_content.y_offset is None
        assert static_content.elapsed_time is None
        assert static_content.total_duration == 50
        assert static_content.extra_time_after_end is None

    def test_update_no_reset(self):
        static_content = movement.StaticContent(0)
        with pytest.raises(movement.MovementException):
            static_content.update(20)

    def test_reset(self):
        static_content = movement.StaticContent(10, 6, 7)
        static_content.set_offset(3, 4)
        static_content.reset()
        assert static_content.x_offset == 6
        assert static_content.y_offset == 7
        assert static_content.total_duration == 10
        assert static_content.elapsed_time == 0
        assert static_content.extra_time_after_end == 0
        assert not static_content.end_of_segment_reached

    def test_forever(self):
        static_content = movement.StaticContent(0, 2, 3)
        static_content.reset()
        assert not static_content.update(13)
        assert not static_content.end_of_segment_reached
        assert static_content.x_offset == 2
        assert static_content.y_offset == 3
        assert static_content.total_duration == 0
        assert static_content.elapsed_time == 0
        assert static_content.extra_time_after_end == 0

    def test_end_not_reached(self):
        static_content = movement.StaticContent(10, 2, 3)
        static_content.reset()
        assert not static_content.update(3)
        assert not static_content.end_of_segment_reached
        assert static_content.x_offset == 2
        assert static_content.y_offset == 3
        assert static_content.total_duration == 10
        assert static_content.elapsed_time == 3
        assert static_content.extra_time_after_end == 0

    def test_end_reached(self):
        static_content = movement.StaticContent(10, 2, 3)
        static_content.reset()
        assert not static_content.update(23)
        assert static_content.end_of_segment_reached
        assert static_content.x_offset == 2
        assert static_content.y_offset == 3
        assert static_content.total_duration == 10
        assert static_content.elapsed_time == 10
        assert static_content.extra_time_after_end == 13

    def test_at_1_before_total_duration_add_1(self):
        static_content = movement.StaticContent(10, 2, 3)
        static_content.reset()
        assert not static_content.update(9)  # 1 before end
        assert not static_content.end_of_segment_reached

        assert not static_content.update(1)  # add 1
        assert static_content.end_of_segment_reached  # this has changed
        assert static_content.extra_time_after_end == 0

        assert not static_content.update(1)  # add 1
        assert static_content.end_of_segment_reached  # still True
        assert static_content.extra_time_after_end == 1

    def test_set_offset(self):
        static_content = movement.StaticContent(10, 2, 3)
        static_content.reset()
        static_content.set_offset(4, 5)
        assert static_content.x_offset == 4
        assert static_content.y_offset == 5


class TestScrollingContent:
    def test_init(self):
        scrolling_content = movement.ScrollingContent(30, 4, -1, 0, 2, 3)
        assert scrolling_content.x_offset_init == 2
        assert scrolling_content.y_offset_init == 3
        assert scrolling_content.x_offset is None
        assert scrolling_content.y_offset is None
        assert scrolling_content.current_step_num is None
        assert scrolling_content.elapsed_time is None
        assert scrolling_content.extra_time_after_end is None
        assert scrolling_content.total_duration == 120
        assert scrolling_content.step_duration == 30
        assert scrolling_content.total_steps == 4
        assert scrolling_content.delta_x_per_step == -1
        assert scrolling_content.delta_y_per_step == 0

    def test_update_no_reset(self):
        scrolling_content = movement.ScrollingContent(30, 4, -1, 0)
        with pytest.raises(movement.MovementException):
            scrolling_content.update(20)

    def test_reset(self):
        scrolling_content = movement.ScrollingContent(30, 4, -1, 0, 2, 3)
        scrolling_content.reset()
        assert scrolling_content.x_offset == 1  # moves at t=0
        assert scrolling_content.y_offset == 3
        assert scrolling_content.current_step_num == 0
        assert scrolling_content.elapsed_time == 0
        assert scrolling_content.extra_time_after_end == 0

    def test_perpetual_scroll(self):
        scrolling_content = movement.ScrollingContent(30, 0, 0, -1, 2, 3)
        scrolling_content.reset()
        assert scrolling_content.update(130)
        assert not scrolling_content.end_of_segment_reached
        assert scrolling_content.x_offset == 2
        assert scrolling_content.y_offset == -2  # moves at t=0, 30, 60, 90, 120
        assert scrolling_content.total_duration == 0
        assert scrolling_content.step_duration == 30
        assert scrolling_content.total_steps == 0
        assert scrolling_content.current_step_num == 0
        assert scrolling_content.elapsed_time == 10
        assert scrolling_content.extra_time_after_end == 0

    def test_within_step_scroll_left(self):
        scrolling_content = movement.ScrollingContent(30, 4, -1, 0, 2, 3)
        scrolling_content.reset()
        assert not scrolling_content.update(20)
        assert not scrolling_content.end_of_segment_reached
        assert scrolling_content.x_offset == 1
        assert scrolling_content.y_offset == 3
        assert scrolling_content.current_step_num == 0
        assert scrolling_content.elapsed_time == 20
        assert scrolling_content.extra_time_after_end == 0

    def test_across_steps_scroll_down(self):
        scrolling_content = movement.ScrollingContent(30, 4, 0, 1, 2, 3)
        scrolling_content.reset()
        assert not scrolling_content.update(20)
        assert not scrolling_content.end_of_segment_reached
        assert scrolling_content.update(50)
        assert not scrolling_content.end_of_segment_reached
        assert scrolling_content.x_offset == 2
        assert scrolling_content.y_offset == 6
        assert scrolling_content.current_step_num == 2
        assert scrolling_content.elapsed_time == 70
        assert scrolling_content.extra_time_after_end == 0

    def test_end_reached_scroll_left(self):
        scrolling_content = movement.ScrollingContent(30, 4, -1, 0, 2, 3)
        scrolling_content.reset()
        scrolling_content.update(80)
        assert scrolling_content.update(70)
        assert scrolling_content.end_of_segment_reached
        assert scrolling_content.x_offset == -2
        assert scrolling_content.y_offset == 3
        assert scrolling_content.current_step_num == 3
        assert scrolling_content.elapsed_time == 120
        assert scrolling_content.extra_time_after_end == 30

    def test_at_1_before_total_duration_add_1(self):
        scrolling_content = movement.ScrollingContent(30, 4, 0, 1, 2, 3)
        scrolling_content.reset()
        # total duration == 120
        assert scrolling_content.update(119)  # 1 before end
        assert not scrolling_content.end_of_segment_reached
        assert scrolling_content.x_offset == 2
        assert scrolling_content.y_offset == 7
        assert scrolling_content.current_step_num == 3

        # add 1
        assert not scrolling_content.update(1)  # no changed as end reached
        assert scrolling_content.end_of_segment_reached  # this has changed
        assert scrolling_content.x_offset == 2
        assert scrolling_content.y_offset == 7  # but not this
        assert scrolling_content.current_step_num == 3  # nor this
        assert scrolling_content.elapsed_time == 120
        assert scrolling_content.extra_time_after_end == 0

        # add 1
        assert not scrolling_content.update(1)
        assert scrolling_content.end_of_segment_reached  # no change
        assert scrolling_content.x_offset == 2
        assert scrolling_content.y_offset == 7  # no change
        assert scrolling_content.current_step_num == 3  # no change
        assert scrolling_content.elapsed_time == 120
        assert scrolling_content.extra_time_after_end == 1  # change

    def test_reset_scroll(self):
        scrolling_content = movement.ScrollingContent(30, 4, 0, -1, 2, 3)
        scrolling_content.reset()
        scrolling_content.update(80)
        scrolling_content.update(70)
        # (assume other movements happen, then loop back to this movement)
        scrolling_content.reset()
        scrolling_content.set_offset(8, 9)
        assert not scrolling_content.end_of_segment_reached
        assert scrolling_content.x_offset == 8
        assert scrolling_content.y_offset == 9
        assert scrolling_content.current_step_num == 0
        assert scrolling_content.elapsed_time == 0
        assert scrolling_content.extra_time_after_end == 0

    def test_set_offset(self):
        scrolling_content = movement.ScrollingContent(30, 4, -1, 0, 2, 3)
        scrolling_content.reset()
        scrolling_content.update(40)  # at t==40
        assert scrolling_content.x_offset == 0
        assert scrolling_content.y_offset == 3

        scrolling_content.set_offset(6, 8)
        scrolling_content.update(60)  # at t==100
        assert scrolling_content.x_offset == 4
        assert scrolling_content.y_offset == 8

    def test_set_offset_after_end(self):
        scrolling_content = movement.ScrollingContent(30, 4, -1, 0, 2, 3)
        scrolling_content.reset()
        scrolling_content.update(150)
        assert scrolling_content.x_offset == -2
        assert scrolling_content.y_offset == 3
        scrolling_content.set_offset(6, 8)
        scrolling_content.update(60)
        # no further movement
        assert scrolling_content.x_offset == 6
        assert scrolling_content.y_offset == 8


class TestMovementCycle:  # pylint: disable=attribute-defined-outside-init
    def setup_method(self):
        self.tracker = movement.MovementCycle(
            [movement.StaticContent(200), movement.ScrollingContent(15, 4, -1, 0)]
        )

    def test_init(self):
        assert len(self.tracker.anim_sequence) == 2
        assert isinstance(self.tracker.anim_sequence[0], movement.StaticContent)
        assert isinstance(self.tracker.anim_sequence[1], movement.ScrollingContent)
        assert not self.tracker._started

        assert self.tracker.cur_segment is None
        assert self.tracker.anim_sequence[0].x_offset is None
        assert self.tracker.anim_sequence[0].y_offset is None
        assert self.tracker.anim_sequence[0].extra_time_after_end is None
        assert self.tracker.anim_sequence[0].elapsed_time is None
        assert self.tracker.anim_sequence[1].x_offset is None
        assert self.tracker.anim_sequence[1].y_offset is None
        assert self.tracker.anim_sequence[1].extra_time_after_end is None
        assert self.tracker.anim_sequence[1].elapsed_time is None
        assert self.tracker.x_offset is None
        assert self.tracker.y_offset is None

    def test_reset(self):
        self.tracker.reset()
        self.tracker.set_offset(2, 3)
        assert self.tracker._started
        assert self.tracker.cur_segment == 0
        assert self.tracker.anim_sequence[0].x_offset == 2
        assert self.tracker.anim_sequence[0].y_offset == 3
        assert self.tracker.anim_sequence[0].extra_time_after_end == 0
        assert self.tracker.anim_sequence[0].elapsed_time == 0
        assert self.tracker.x_offset == 2
        assert self.tracker.y_offset == 3

    def test_within_segment(self):
        self.tracker.reset()
        self.tracker.set_offset(2, 3)
        self.tracker.update(180)
        assert self.tracker.cur_segment == 0
        assert self.tracker.anim_sequence[0].x_offset == 2
        assert self.tracker.anim_sequence[0].y_offset == 3
        assert self.tracker.anim_sequence[0].extra_time_after_end == 0
        assert self.tracker.anim_sequence[0].elapsed_time == 180
        assert not self.tracker.anim_sequence[0].end_of_segment_reached
        assert self.tracker.x_offset == 2
        assert self.tracker.y_offset == 3

    def test_across_segments(self):
        self.tracker.reset()
        self.tracker.set_offset(2, 3)
        self.tracker.update(235)
        assert self.tracker.cur_segment == 1
        assert self.tracker.anim_sequence[0].extra_time_after_end == 35
        assert self.tracker.anim_sequence[0].elapsed_time == 200
        assert self.tracker.anim_sequence[0].end_of_segment_reached
        assert self.tracker.anim_sequence[0].x_offset == 2
        assert self.tracker.anim_sequence[0].y_offset == 3
        assert self.tracker.anim_sequence[1].x_offset == -1
        assert self.tracker.anim_sequence[1].y_offset == 3
        assert self.tracker.anim_sequence[1].extra_time_after_end == 0
        assert self.tracker.anim_sequence[1].elapsed_time == 35
        assert self.tracker.anim_sequence[1].current_step_num == 2
        assert not self.tracker.anim_sequence[1].end_of_segment_reached
        assert self.tracker.x_offset == -1
        assert self.tracker.y_offset == 3

    def test_full_cycle(self):
        self.tracker.reset()
        self.tracker.set_offset(2, 3)
        self.tracker.update(500)
        assert self.tracker.cur_segment == 1
        assert self.tracker.anim_sequence[0].extra_time_after_end == 40
        assert self.tracker.anim_sequence[0].elapsed_time == 200
        assert self.tracker.anim_sequence[0].end_of_segment_reached
        assert self.tracker.anim_sequence[1].x_offset == -5
        assert self.tracker.anim_sequence[1].y_offset == 3
        assert self.tracker.anim_sequence[1].extra_time_after_end == 0
        assert self.tracker.anim_sequence[1].elapsed_time == 40
        assert self.tracker.anim_sequence[1].current_step_num == 2
        assert not self.tracker.anim_sequence[1].end_of_segment_reached
        assert self.tracker.x_offset == -5
        assert self.tracker.y_offset == 3

    def test_set_offset(self):
        self.tracker.reset()
        self.tracker.set_offset(2, 3)
        self.tracker.update(500)  # init with output of previous test
        assert self.tracker.x_offset == -5
        assert self.tracker.y_offset == 3
        self.tracker.set_offset(8, 2)
        assert self.tracker.x_offset == 8
        assert self.tracker.y_offset == 2
        self.tracker.update(15)
        assert self.tracker.cur_segment == 1
        assert self.tracker.anim_sequence[1].x_offset == 7
        assert self.tracker.anim_sequence[1].y_offset == 2
        assert self.tracker.anim_sequence[1].elapsed_time == 55
        assert self.tracker.anim_sequence[1].current_step_num == 3



def rchop(s, ending):
    return s[: -len(ending)] if s.endswith(ending) else s


def lchop(s, beginning):
    return s[len(beginning) :] if s.startswith(beginning) else s


def ordinal_en(n: int):
    # https://stackoverflow.com/questions/9647202/ordinal-numbers-replacement
    return f'{n}{"tsnrhtdd"[(n//10%10!=1)*(n%10<4)*n%10::4]}'


def ordinal_fr(n: int):
    if n == 1:
        return "1er"
    return f"{n}è"


def arrival_time_en(time_in_seconds: int):
    if time_in_seconds == 0:
        return "Here"

    time_in_mins = time_in_seconds // 60
    if time_in_mins == 0:
        return ""

    return str(time_in_mins) + (" min" if time_in_mins == 1 else " mins")



"""
UI for Nederlandse Spoorwegen
"""

from tabulate import tabulate

import departure.commons.helpers as helpers
from . import data


def list_departures_with_schedule(
    departures_with_schedule: list, current_station_ns_code: str
) -> None:
    """List departures with schedule.

    Args:
        departures_with_schedule (list): departures with schedule, as returned by
            ns.departures_with_schedule
        current_station_ns_code (str): NS code of the current station, used to only
            show stops after the current station
    """
    for i, train in enumerate(departures_with_schedule):
        print(
            f'{helpers.ordinal_en(i+1)} {train["train_number"]} '
            f'{train["direction"]} ',
            end="",
        )
        if train["actual_time"] == train["planned_time"]:
            print(f'{train["actual_time"]}', end="")
        else:
            print(f'*{train["actual_time"]} ({train["planned_time"]})', end="")
        print(f' ({train["train_category"]})')

        if "schedule" in train:
            print_schedule(train["schedule"], current_station_ns_code)


def print_schedule(schedule: list, current_station_ns_code: str):
    stations_by_uic = data.STATIONS_BY_UIC
    current_station_uic = data.STATIONS_BY_NS_CODE[current_station_ns_code]["uic"]
    before_current_station = True

    stops = []
    for stop in schedule:
        # ignore stops before current station
        if stop["stop_uic"] == current_station_uic:
            before_current_station = False
            continue

        if before_current_station:
            continue

        # print schedule after current station
        if stop["stop_uic"] in stations_by_uic:
            stop_name = stations_by_uic[stop["stop_uic"]]["name_medium"]
        else:  # don't attempt to resolve using UIC if stop is outside NL
            stop_name = f"{stop['stop_name']}"
        stops.append(f"{stop_name} ({stop['actual_time']} / {stop['planned_time']})")

    print(f"  Calling at: {', '.join(stops)}")


def list_stations(stations: list):
    print(
        tabulate(
            [
                [
                    station_id,
                    stations[station_id]["uic"],
                    stations[station_id]["name_medium"],
                    stations[station_id]["name_long"],
                ]
                for station_id in sorted(
                    stations, key=lambda i: stations[i]["name_long"]
                )
            ],
            headers=["NS code", "UIC", "name (medium)", "name (long)"],
        )
    )

    print(f"> total: {len(stations)}")


def print_station_by_ns_code(ns_code: str):
    stations = data.STATIONS_BY_NS_CODE

    if ns_code in stations:
        print(f"{stations[ns_code]['name_long']} ({stations[ns_code]['uic']})")
    else:
        print(f"! invalid NS code: {ns_code}")



"""
UI for TfL tube
"""

from tabulate import tabulate

import departure.commons.helpers as helpers
from . import data


def expanded_station_id(station_id: str) -> str:
    station_id = station_id.upper()
    if len(station_id) == 3:
        return "940GZZLU" + station_id
    return station_id


def expanded_direction(direction: str) -> str:
    direction = direction.lower()
    return {
        "w": "Westbound",
        "e": "Eastbound",
        "s": "Southbound",
        "n": "Northbound",
        "i": "Inner Rail",
        "o": "Outer Rail",
    }.get(direction[0], direction)


def expanded_line_id(line_id: str) -> str:
    line_id = line_id.lower()
    if line_id[0] == "c":
        return {
            "ce": "central",
            "ci": "circle",
        }.get(line_id[:2], line_id)
    return {
        "b": "bakerloo",
        "d": "district",
        "h": "hammersmith-city",
        "j": "jubilee",
        "m": "metropolitan",
        "n": "northern",
        "p": "piccadilly",
        "v": "victoria",
        "w": "waterloo-city",
    }.get(line_id[0], line_id)


def list_lines() -> None:
    table = []
    for line_id in data.lines:
        for canonical_direction in data.lines[line_id]["canonical_directions"]:
            table.append(
                [
                    line_id,
                    canonical_direction,
                    data.lines[line_id]["canonical_directions"][canonical_direction],
                ]
            )

    print(tabulate(table, headers=["line id", "direction", "direction (alt.)"]))


def list_stations(stations=None) -> None:
    # default: list all stations
    if not stations:
        stations = data.STATIONS

    print(
        tabulate(
            [
                [
                    station_id,
                    stations[station_id]["name"],
                    " ".join(stations[station_id]["lines"]),
                ]
                for station_id in sorted(stations)
            ],
            headers=["station id", "station name", "line ids"],
        )
    )

    print(f"> total: {len(stations)}")


def list_trains_single_station(trains: dict):
    table = []
    i = 1

    # populate table
    for train in sorted(trains.values(), key=lambda a: a["time_to_station"]):
        table.append(
            [
                helpers.ordinal_en(i),
                train["towards"],
                helpers.arrival_time_en(train["time_to_station"]),
            ]
        )
        i += 1

    print(tabulate(table, tablefmt="plain"))


def list_trains_multiple_stations(trains: dict):
    # reorganise by station
    trains_by_station = {}

    for train_id in trains:
        station_id = trains[train_id]["station_id"]
        if station_id not in trains_by_station:
            trains_by_station[station_id] = {}
        trains_by_station[station_id][train_id] = trains[train_id]

    # list by station
    for station_id in trains_by_station:
        print(station_id)
        list_trains_single_station(trains_by_station[station_id])
        print()



import time
import threading
import logging

import departure.board.view_model as view_model

logger = logging.getLogger(__name__)


class DataUpdater(threading.Thread):
    def __init__(
        self,
        target_view_model: view_model.ViewModel,
        data_refresh_delay_in_s: int,
        end_event: threading.Event,
    ):
        super().__init__()
        self.view_model = target_view_model
        self.data_refresh_delay_in_s = data_refresh_delay_in_s
        self.end_event = end_event

    def run(self):
        while not self.end_event.is_set():
            loop_start_time = time.monotonic()
            next_loop_start_time = loop_start_time + self.data_refresh_delay_in_s

            self.update()

            if next_loop_start_time > time.monotonic():
                self.end_event.wait(next_loop_start_time - time.monotonic())
            else:
                logger.warning("exceeded data update refresh delay")

        logger.info("data updater stopped running")

    def update(self):
        raise NotImplementedError()



from typing import List, Union

import departure.board.departure_pb2 as departure_pb2
import departure.board.board as board
import departure.board.movement as movement


def deserialise_Pixel(pixel: departure_pb2.Pixel):
    return pixel.x, pixel.y, (pixel.r, pixel.g, pixel.b)


def deserialise_BoardSectionContent(
    board_section_content: departure_pb2.BoardSectionContent,
) -> board.BoardSectionContent:
    return board.BoardSectionContent(
        content_pixels=[
            deserialise_Pixel(pixel) for pixel in board_section_content.pixels
        ],
        content_w=board_section_content.content_w,
        content_h=board_section_content.content_h,
        repeat_x=board_section_content.repeat_x,
        repeat_y=board_section_content.repeat_y,
    )


def deserialise_NoMovement(
    no_movement: departure_pb2.NoMovement,
) -> movement.NoMovement:
    return movement.NoMovement(
        x_offset_init=no_movement.x_offset_init, y_offset_init=no_movement.y_offset_init
    )


def deserialise_StaticContent(
    static_content: departure_pb2.StaticContent,
) -> movement.StaticContent:
    return movement.StaticContent(
        total_duration=static_content.total_duration,
        x_offset_init=static_content.x_offset_init,
        y_offset_init=static_content.y_offset_init,
    )


def deserialise_ScrollingContent(
    scrolling_content: departure_pb2.ScrollingContent,
) -> movement.ScrollingContent:
    return movement.ScrollingContent(
        step_duration=scrolling_content.step_duration,
        total_steps=scrolling_content.total_steps,
        delta_x_per_step=scrolling_content.delta_x_per_step,
        delta_y_per_step=scrolling_content.delta_y_per_step,
        x_offset_init=scrolling_content.x_offset_init,
        y_offset_init=scrolling_content.y_offset_init,
    )


def deserialise_Movement(
    single_movement: departure_pb2.Movement,
) -> Union[movement.Movement, None]:
    if single_movement.HasField("no_movement"):
        return deserialise_NoMovement(single_movement.no_movement)

    if single_movement.HasField("static_content"):
        return deserialise_StaticContent(single_movement.static_content)

    if single_movement.HasField("scrolling_content"):
        return deserialise_ScrollingContent(single_movement.scrolling_content)

    return None


def deserialise_RepeatedMovement(
    movement_list: List[departure_pb2.Movement],
) -> Union[None, movement.Movement]:
    # no movement
    if movement_list is None or len(movement_list) == 0:
        return None

    # normal movement
    if len(movement_list) == 1:
        return deserialise_Movement(movement_list[0])

    # movement cycle
    movement_segments = []
    for current_movement in movement_list:
        movement_segment = deserialise_Movement(current_movement)

        # invalid case
        if movement_segment is None:
            return None

        movement_segments.append(movement_segment)

    return movement.MovementCycle(movement_segments)


def serialise_pixels(pixels) -> List[departure_pb2.Pixel]:
    serialised_pixels = []
    for x, y, colour in pixels:
        serialised_pixels.append(
            departure_pb2.Pixel(x=x, y=y, r=colour[0], g=colour[1], b=colour[2])
        )
    return serialised_pixels



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from departure.board import departure_pb2 as departure_dot_board_dot_departure__pb2


class BoardManagerStub(object):
    """service"""

    def __init__(self, channel):
        """Constructor.

        Args:
          channel: A grpc.Channel.
        """
        self.BoardSectionsUpdate = channel.unary_unary(
            "/departure.proto.BoardManager/BoardSectionsUpdate",
            request_serializer=departure_dot_board_dot_departure__pb2.BoardSectionsUpdateRequest.SerializeToString,
            response_deserializer=departure_dot_board_dot_departure__pb2.BoardSectionsUpdateResponse.FromString,
        )


class BoardManagerServicer(object):
    """service"""

    def BoardSectionsUpdate(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_BoardManagerServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "BoardSectionsUpdate": grpc.unary_unary_rpc_method_handler(
            servicer.BoardSectionsUpdate,
            request_deserializer=departure_dot_board_dot_departure__pb2.BoardSectionsUpdateRequest.FromString,
            response_serializer=departure_dot_board_dot_departure__pb2.BoardSectionsUpdateResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "departure.proto.BoardManager", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))



"""
CLI for London Underground (TfL Tube)
"""

import threading
import time
import logging
import cProfile
import pstats

import click

from departure.board import board_client

from . import ui, tfl_tube, commons, view_model, data_updater
from . import api

ONE_DAY_IN_SECONDS = 60 * 60 * 24
logger = logging.getLogger(__name__)


@click.group()
def cli():
    """
    Information and departures for London Underground (UK).

    Note - Your TfL application key must be assigned to the TFL_APP_KEY environment
    variable.
    """


@click.command(name="search")
@click.argument("query_string", required=False, default="")
def search(query_string):
    """Search for stations."""
    if query_string:
        stations_result = tfl_tube.stations_by_string(query_string)
        ui.list_stations(stations_result)
    else:
        ui.list_stations()


@click.command()
def lines():
    """List lines and directions."""
    ui.list_lines()


@click.command(name="next", short_help="Get next services at station.")
@click.argument("line_id")
@click.argument("station_id")
@click.argument("direction")
@click.option(
    "--profiling", is_flag=True, help="(debugging) performance profiling"
)
def next_departures(line_id, station_id, direction, profiling):
    """
    Get next departures at station STATION_ID, on line LINE_ID, in direction
    DIRECTION.

    Use the lines command to find the LINE_ID and possible DIRECTIONs for a line,
    and the search command to find the STATION_ID and LINE_ID for a station.

    Note - This command accepts short versions of its argument values: the line's
    initial (or first two characters for the Central and Circle lines), the last three
    characters of the station, and the direction's initial.
    For example you can use 'next ce oxc w' to find the next Westbound departures on
    the Central line at Oxford Circus.
    """

    try:
        if profiling:
            prof = cProfile.Profile()
            prof.enable()

        trains = tfl_tube.next_trains(
            ui.expanded_line_id(line_id),
            ui.expanded_station_id(station_id),
            ui.expanded_direction(direction),
        )
        ui.list_trains_single_station(trains)

        if profiling:
            prof.disable()
            prof_stats = pstats.Stats(prof)
            prof_stats.strip_dirs()
            prof_stats.sort_stats("cumulative")
            prof_stats.print_stats("tfl_tube")

    except commons.TflTubeException as e:
        print(f"error: {str(e)}")


@click.command(name="at-station", help="(debugging) list arrivals")
@click.option("--line_ids", help="LINE_ID[,LINE_ID]*")
@click.option("--station_id", help="STATION_ID")
@click.option("--direction", help="DIRECTION")
def at_station(line_ids=None, station_id=None, direction=None):
    if line_ids is not None:
        line_ids = line_ids.split(",")
        line_ids = [ui.expanded_line_id(l) for l in line_ids]

    if station_id is not None:
        station_id = ui.expanded_station_id(station_id)

    if direction is not None:
        direction = ui.expanded_direction(direction)

    try:
        arrivals = api.tube_arrivals(count=-1)

        if arrivals is None:
            print("no arrivals")
            return

        trains = tfl_tube.filter_trains_currently_at_platform(
            arrivals, line_ids, station_id, direction
        )
        ui.list_trains_multiple_stations(trains)
    except commons.TflTubeException as e:
        print(f"error: {str(e)}")


@click.command(name="board")
@click.argument("line_id")
@click.argument("station_id")
@click.argument("direction_id")
def start_board(line_id, station_id, direction_id):
    """Update a departure board with departures at station STATION_ID, on line
    LINE_ID in direction DIRECTION_CODE.

    Use the lines command to find the LINE_ID and possible DIRECTIONs for a line,
    and the search command to find the STATION_ID and LINE_ID for a station.

    Note - This command accepts short versions of its argument values: the line's
    initial (or first two characters for the Central and Circle lines), the last three
    characters of the station, and the direction's initial.
    For example you can use 'next ce oxc w' to find the next Westbound departures on
    the Central line at Oxford Circus.
    """

    line_id = ui.expanded_line_id(line_id)
    station_id = ui.expanded_station_id(station_id)
    direction = ui.expanded_direction(direction_id)

    try:
        tfl_tube.check_params([line_id], station_id, direction)
    except commons.TflTubeException as e:
        logger.error(str(e))
        return

    # start board
    board_client_instance = board_client.BoardClient()

    threading.Thread(
        target=board_client_instance.run,
        args=[
            view_model.ViewModelTflTube_192_32_3_Rows_To_ProtocolBuffers,
            data_updater.DataUpdaterTflTube,
            {"line_id": line_id, "station_id": station_id, "direction": direction},
        ],
    ).start()

    # run until interrupt
    try:
        while True:
            time.sleep(ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        logger.info("received interrupt")  # debugging
        board_client_instance.running = False


cli.add_command(search)
cli.add_command(lines)
cli.add_command(next_departures)
cli.add_command(at_station)
cli.add_command(start_board)

if __name__ == "__main__":
    cli()



"""
CLI for RATP
"""

import threading
import time
import logging

import click

from departure.board import board_client

from . import ui, ratp, view_model, data_updater

ONE_DAY_IN_SECONDS = 60 * 60 * 24
logger = logging.getLogger(__name__)


@click.group()
def cli():
    """
    Information and departures for RATP (FR).
    """


@click.command()
@click.argument("query_string")
def search(query_string):
    """Search for stations containing QUERY_STRING."""
    stations = ratp.stations_by_name(query_string)
    ui.list_stations(stations, show_line=True, show_station_id=True)


@click.command(name="lines")
@click.argument("code_query", required=False, default="")
def lines_by_code(code_query):
    """
    Search for lines by number.

    If CODE_QUERY is provided, search line numbers containing the query text.
    Otherwise, list all lines.
    """
    lines_result = ratp.lines_by_code(code_query)
    ui.list_lines(lines_result)


@click.command(name="lines-name")
@click.argument("line_query", required=False, default="")
def lines_by_name(line_query):
    """
    Search for lines by name (termini).

    If LINE_QUERY is provided, search line names containing the query text.
    Otherwise, list all lines.
    """
    lines = ratp.lines_by_name(line_query.lower())
    ui.list_lines(lines)


@click.command()
@click.argument("line_id")
def directions(line_id):
    """
    Get directions for LINE_ID.

    Use line search commands to find the LINE_ID for a line.
    """

    line_directions = ratp.directions(line_id)
    ui.list_directions(line_directions)


@click.command(name="stations-line")
@click.argument("line_id")
def stations_by_line(line_id):
    """
    Get stations on LINE_ID.

    Use line search commands to find the LINE_ID for a line.
    """

    stations = ratp.stations_by_line(line_id)
    ui.list_stations(stations, show_station_id=True)


@click.command(name="next", short_help="Get next departures at station.")
@click.argument("line_id")
@click.argument("line_station_id")
@click.argument("direction")
def next_departures(line_id, line_station_id, direction):
    """
    Get next departures at station LINE_STATION_ID, on line LINE_ID, in direction
    DIRECTION.

    Use the line search commands to find the LINE_ID for a line, the stations-line
    command to find the LINE_STATION_ID for a station on a specific line, and the
    direction command to find the DIRECTIONs for a line.
    """

    departures = ratp.next_departures(line_id, line_station_id, direction)
    ui.list_departures(departures)


@click.command(
    name="board", short_help="Update a departure board with departures at a station."
)
@click.argument("line_id")
@click.argument("line_station_id")
@click.argument("direction_code")
def start_board(line_id, line_station_id, direction_code):
    """Update a departure board with departures at station LINE_STATION_ID, on line
    LINE_ID in direction DIRECTION_CODE.

    Use the line search commands to find the LINE_ID for a line, the stations-line
    command to find the LINE_STATION_ID for a station on a specific line, and the
    direction command to find the DIRECTION_CODEs for a line.
    """

    # start board
    board_client_instance = board_client.BoardClient()

    threading.Thread(
        target=board_client_instance.run,
        args=[
            view_model.ViewModelRatp_192_32_3_Rows_To_ProtocolBuffers,
            data_updater.DataUpdaterRatp,
            {
                "line_id": line_id,
                "line_station_id": line_station_id,
                "direction": direction_code,
            },
        ],
    ).start()

    # run until interrupt
    try:
        while True:
            time.sleep(ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        logger.info("received interrupt")
        board_client_instance.running = False


cli.add_command(search)
cli.add_command(lines_by_code)
cli.add_command(lines_by_name)
cli.add_command(directions)
cli.add_command(stations_by_line)
cli.add_command(next_departures)
cli.add_command(start_board)

if __name__ == "__main__":
    cli()



import xml.etree.ElementTree as ET
import os
import logging

import requests

from . import commons

logger = logging.getLogger(__name__)


def api_request(url: str):
    commons.check_env_vars()

    try:
        response = requests.get(
            url,
            timeout=15,
            auth=(os.environ["TRANSILIEN_USER"], os.environ["TRANSILIEN_PASSWORD"]),
        )
    except requests.exceptions.Timeout:
        logger.warning("Transilien API HTTP request timed out")
        return None
    except Exception as e:  # pylint: disable=broad-except
        logger.warning("Transilien API HTTP request error: %s", str(e))
        return None

    # process response
    if response.status_code == 200:
        return ET.fromstring(response.content.decode("utf-8"))

    return None


def departures(station_id):
    url = f"https://api.transilien.com/gare/{station_id}/depart/"
    return api_request(url)



import threading
import os
import logging

import grpc

import departure.board.departure_pb2_grpc as departure_pb2_grpc

from . import data_updater, view_model

logger = logging.getLogger(__name__)


class BoardClient:  # pylint: disable=too-few-public-methods
    def __init__(self):
        self.running = False

    def run(
        self,
        view_model_class: view_model.ViewModel,
        data_updater_class: data_updater.DataUpdater,
        data_updater_extra_arguments: dict,
    ) -> None:

        end_event = threading.Event()

        # set departure board server
        if "DEPARTURE_BOARD_SERVER" in os.environ:
            departure_board_server = os.environ["DEPARTURE_BOARD_SERVER"]
            server_info_str = "DEPARTURE_BOARD_SERVER env var set - "
        else:
            server_info_str = "DEPARTURE_BOARD_SERVER env var not set - "
            departure_board_server = "127.0.0.1"
        logger.info("%s using %s as server", server_info_str, departure_board_server)

        # open channel to board server
        channel = grpc.insecure_channel(f"{departure_board_server}:50051")

        # set up client parameters
        board_manager_stub = departure_pb2_grpc.BoardManagerStub(channel)
        target_view_model = view_model_class(board_manager_stub)

        # init and run updater thread
        data_updater_thread = data_updater_class(
            target_view_model=target_view_model,
            data_refresh_delay_in_s=15,
            end_event=end_event,
            **data_updater_extra_arguments,
        )

        data_updater_thread.start()

        self.running = True

        # run until self.running is set to false
        while self.running:
            end_event.wait(0.5)

        # clean and exit
        logger.info("stopping")
        end_event.set()
        data_updater_thread.join()
        target_view_model.update(None)



"""
CLI for SNCF
"""

import threading
import time
import logging

import click

from departure.board import board_client

from . import ui, sncf, commons, view_model, data_updater

ONE_DAY_IN_SECONDS = 60 * 60 * 24
logger = logging.getLogger(__name__)


@click.group()
def cli():
    """
    Information and departures for SNCF (FR).

    Note - Your SNCF authorisation key must be assigned to the SNCF_KEY environment
    variable.
    """


@click.command()
@click.argument("query_string")
def search(query_string):
    """Search for stations containing QUERY_STRING."""
    stations = sncf.stations_by_string(query_string)
    ui.list_stations(stations)


@click.command(name="next")
@click.argument("stop_area_id")
@click.option(
    "--full",
    "timetable_for_all_trains",
    is_flag=True,
    help="displays timetable for all trains, instead of just first one",
)
def next_departures(stop_area_id, timetable_for_all_trains):
    """
    Get next departures at station STOP_AREA_ID.

    Use search to find the STOP_AREA_ID for a station.
    """

    trains = sncf.next_trains(stop_area_id, timetable_for_all_trains)
    ui.list_trains(trains)


@click.command(name="board")
@click.argument("stop_area_id")
def start_board(stop_area_id):
    """Update a departure board with departures at station STOP_AREA_ID.

    Use search to find the STOP_AREA_ID for a station."""

    try:
        sncf.check_params(stop_area_id)
    except commons.SncfException as e:
        logger.error("error: %s", str(e))

    # start board
    board_client_instance = board_client.BoardClient()

    threading.Thread(
        target=board_client_instance.run,
        args=[
            view_model.ViewModelSncf_192_32_3_Rows_To_ProtocolBuffers,
            data_updater.DataUpdaterSncf,
            {
                "stop_area_id": stop_area_id,
            },
        ],
    ).start()

    # run until interrupt
    try:
        while True:
            time.sleep(ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        logger.info("received interrupt")
        board_client_instance.running = False


cli.add_command(search)
cli.add_command(next_departures)
cli.add_command(start_board)

if __name__ == "__main__":
    cli()



import threading
import time
import logging

from fastapi import APIRouter
from pydantic import BaseModel
from starlette.requests import Request

from . import tfl_tube, commons, view_model, data_updater

logger = logging.getLogger(__name__)

router = APIRouter()


class StationDirection(BaseModel):
    line_id: str
    station_id: str
    direction: str


@router.get("/stations/{station_string_query}")
async def search_stations(station_string_query: str):
    return tfl_tube.stations_by_string(station_string_query)


@router.get("/lines")
async def lines():
    return tfl_tube.lines()


@router.get("/stations-line/{line_id}")
async def stations_line(line_id: str):
    return tfl_tube.stations_by_line(line_id)


@router.get("/directions/{line_id}")
async def directions(line_id: str):
    return tfl_tube.directions(line_id)


# debugging only
@router.get("/next/{line_id}/{station_id}/{direction}")
async def next_trains(line_id: str, station_id: str, direction: str):
    try:
        departures = tfl_tube.next_trains(line_id, station_id, direction)
    except commons.TflTubeException as e:
        logger.warning(str(e))
        return {"status": "error", "message": str(e)}

    return {"status": "OK", "response": departures}


# curl -X POST http://localhost:8000/tfl-tube/start-client \
#   -d {\"line_id\":\"district\", \"station_id\":\"940GZZLUVIC\", \"direction\":\"Westbound\"}
@router.post("/start-client")
async def start_client(station_direction: StationDirection, request: Request):
    # check params
    try:
        tfl_tube.check_params(
            [station_direction.line_id],
            station_direction.station_id,
            station_direction.direction,
        )
        commons.check_env_vars()
    except commons.TflTubeException as e:
        logger.warning(str(e))
        return {"status": "error", "message": str(e)}

    # stop board client if already running
    if request.app.board_client.running:
        request.app.board_client.running = False
        time.sleep(1)

    # start board
    threading.Thread(
        target=request.app.board_client.run,
        args=[
            view_model.ViewModelTflTube_192_32_3_Rows_To_ProtocolBuffers,
            data_updater.DataUpdaterTflTube,
            {
                "line_id": station_direction.line_id,
                "station_id": station_direction.station_id,
                "direction": station_direction.direction,
            },
        ],
    ).start()

    return {"status": "OK"}



import threading
import time
import logging

from fastapi import APIRouter
from pydantic import BaseModel
from starlette.requests import Request

from . import sncf, commons, view_model, data_updater

logger = logging.getLogger(__name__)

router = APIRouter()


class Station(BaseModel):
    stop_area_id: str


@router.get("/search/{search_string}")
async def search(search_string):
    return sncf.stations_by_string(search_string)


# debugging only
@router.get("/next/{stop_area_id}")
async def next_trains(stop_area_id):
    stop_area_id = stop_area_id.upper()

    try:
        response = sncf.next_trains(stop_area_id)
    except commons.SncfException as e:
        return {"status": "error", "message": str(e)}

    # request next services
    return {"status": "OK", "response": response}


# curl -X POST http://localhost:8000/sncf/start-client -d {\"stop_area_id\":\"87393702\"}
@router.post("/start-client")
async def start_client(station: Station, request: Request):
    # check parameters
    stop_area_id = station.stop_area_id
    try:
        commons.check_env_vars()
        sncf.check_params(stop_area_id)
    except commons.SncfException as e:
        logger.warning(str(e))
        return {"status": "error", "message": str(e)}

    # stop board client if already running
    if request.app.board_client.running:
        request.app.board_client.running = False
        time.sleep(1)

    # start board
    threading.Thread(
        target=request.app.board_client.run,
        args=[
            view_model.ViewModelSncf_192_32_3_Rows_To_ProtocolBuffers,
            data_updater.DataUpdaterSncf,
            {"stop_area_id": stop_area_id},
        ],
    ).start()

    return {"status": "OK"}



from tabulate import tabulate

from . import data


def list_stations(stations=None):
    # default: list all stations
    if not stations:
        stations = data.STATIONS

    stations_table = []

    for station_id in sorted(stations):
        stations_table.append([station_id, stations[station_id]["nom"]])

    print(tabulate(stations_table, headers=["stop area id", "name"]))
    print(f"> total: {len(stations)}")


def print_timetable(timetable, current_stop_point_id):
    stop_points = []
    before_current_stop_point = True

    for stop_point in timetable:
        # ignore stops before current stop
        if stop_point["id"] == current_stop_point_id:
            before_current_stop_point = False
            continue

        if before_current_stop_point:
            continue

        # print timetable after current stop
        if stop_point["amended_departure_time"] == stop_point["base_departure_time"]:
            stop_points.append(
                f'{stop_point["name"]} ({stop_point["base_departure_time"]})'
            )
        else:
            stop_points.append(
                f'{stop_point["name"]} ({stop_point["amended_departure_time"]} | '
                f'{stop_point["base_departure_time"]})'
            )

    print(f'  Arrêts : {", ".join(stop_points)}')


def list_trains(trains):
    for train in trains:
        print(f'{train["headsign"]} {train["direction"]} ', end="")
        if train["time"] == train["base_time"]:
            print(f'{train["time"]}', end="")
        else:
            print(f'*{train["time"]} ({train["base_time"]})', end="")
        print(f' ({train["commercial_mode"]})')

        if "timetable" in train:
            print_timetable(train["timetable"], train["stop_point_id"])



import departure.board.board as board
import departure.board.movement as movement

# pylint: disable=protected-access


class TestRepeatedOffsets:
    # x_offset, y_offset, display_w, display_h, content_w, content_h

    def test_no_wrap_right_no_repeat(self):
        assert board.repeated_offsets(20, 3, 60, 20, 20, 10) == (
            [20],
            [3],
        )  # right of display_w
        assert board.repeated_offsets(2, 3, 60, 20, 20, 10) == (
            [2],
            [3],
        )  # between 0 and display_w (visible)
        assert board.repeated_offsets(0, 3, 60, 20, 20, 10) == (
            [0],
            [3],
        )  # at limit of visible (visible)
        assert board.repeated_offsets(-1, 3, 60, 20, 20, 10) == (
            [-1],
            [3],
        )  # at limit of visible (non-visible)
        assert board.repeated_offsets(-40, 3, 60, 20, 20, 10) == (
            [-40],
            [3],
        )  # at limit of repeat (non-repeat)

    def test_repeat_right(self):
        assert board.repeated_offsets(-41, 3, 60, 20, 20, 10) == (
            [-41, 19],
            [3],
        )  # at limit of repeat (repeat)
        assert board.repeated_offsets(-59, 3, 60, 20, 20, 10) == (
            [-59, 1],
            [3],
        )  # at limit of wrap (non-wrap)

    def test_no_wrap_down_no_repeat(self):
        assert board.repeated_offsets(2, 20, 60, 20, 20, 10) == (
            [2],
            [20],
        )  # below display_h
        assert board.repeated_offsets(2, 2, 60, 20, 20, 10) == (
            [2],
            [2],
        )  # between 0 and display_h (visible)
        assert board.repeated_offsets(2, 0, 60, 20, 20, 10) == (
            [2],
            [0],
        )  # at limit of visible (visible)
        assert board.repeated_offsets(2, -1, 60, 20, 20, 10) == (
            [2],
            [-1],
        )  # at limit of visible (non-visible)
        assert board.repeated_offsets(2, -10, 60, 20, 20, 10) == (
            [2],
            [-10],
        )  # at limit of repeat (non-repeat)

    def test_repeat_down(self):
        assert board.repeated_offsets(2, -11, 60, 20, 20, 10) == (
            [2],
            [-11, 9],
        )  # at limit of repeat (repeat)
        assert board.repeated_offsets(2, -19, 60, 20, 20, 10) == (
            [2],
            [-19, 1],
        )  # at limit of wrap (non-wrap)


class TestWrapOffset:
    def test_wrap_right(self):
        content = board.BoardSectionContent([], 60, 20)
        move = movement.NoMovement(20, 10)
        section = board.BoardSection(0, 0, 0, 0, content=content, movement=move)

        # at limit of wrap (wrap)
        section._movement.set_offset(-60, 3)
        section._wrap_offset()
        assert (section._movement.x_offset, section._movement.y_offset) == (0, 3)

        # between
        section._movement.set_offset(-61, 3)
        section._wrap_offset()
        assert (section._movement.x_offset, section._movement.y_offset) == (-1, 3)

        # resolves (wrap) to limit of repeat (non-repeat)
        section._movement.set_offset(-100, 3)
        section._wrap_offset()
        assert (section._movement.x_offset, section._movement.y_offset) == (-40, 3)

        # resolves (wrap) to limit of repeat (repeat)
        section._movement.set_offset(-101, 3)
        section._wrap_offset()
        assert (section._movement.x_offset, section._movement.y_offset) == (-41, 3)

    def test_wrap_down(self):
        content = board.BoardSectionContent([], 60, 20)
        move = movement.NoMovement(20, 10)
        section = board.BoardSection(0, 0, 0, 0, content=content, movement=move)

        # at limit of wrap (wrap)
        section._movement.set_offset(2, -20)
        section._wrap_offset()
        assert (section._movement.x_offset, section._movement.y_offset) == (2, 0)

        # between
        section._movement.set_offset(2, -21)
        section._wrap_offset()
        assert (section._movement.x_offset, section._movement.y_offset) == (2, -1)

        # resolves (wrap) to limit of repeat (non-repeat)
        section._movement.set_offset(2, -30)
        section._wrap_offset()
        assert (section._movement.x_offset, section._movement.y_offset) == (2, -10)

        # resolves (wrap) to limit of repeat (repeat)
        section._movement.set_offset(2, -31)
        section._wrap_offset()
        assert (section._movement.x_offset, section._movement.y_offset) == (2, -11)


class TestVisiblePixelsMultipleOffsets:
    # remains marked as not covered by tests due to continue
    # see https://bitbucket.org/ned/coveragepy/issues/198/continue-marked-as-not-covered
    # leaving in case bug gets fixed
    def test_invisible_repetitions(self):
        output_pixels = board.visible_pixels_multiple_offsets(
            [(0, 0, (255, 255, 255))], [0, 8], [0, 8], 8, 8, 8, 8
        )

        assert output_pixels == [(0, 0, (255, 255, 255))]



import threading
import time
import logging

from fastapi import APIRouter
from pydantic import BaseModel
from starlette.requests import Request

from . import ns, commons, view_model, data_updater

logger = logging.getLogger(__name__)

router = APIRouter()


class Station(BaseModel):
    code: str


def check_params(station_code):
    try:
        ns.check_params(station_code)
    except commons.NSException:
        return {"status": "error", "message": "invalid station code"}
    return {}


@router.get("/search/{search_string}")
async def search(search_string):
    return ns.search(search_string)


# debugging only
@router.get("/next/{station_code}")
async def next_departures(station_code):
    # request next departures
    station_code = station_code.upper()

    try:
        response = ns.departures_with_schedule(station_code)
    except commons.NSException as e:
        logger.warning(str(e))
        return {"status": "error", "message": str(e)}

    # request next services
    return {
        "status": "OK",
        "response": response,
    }


# curl -X POST http://localhost:8000/ns/start-client -d {\"code\":\"gvc\"}
@router.post("/start-client")
async def start_client(station: Station, request: Request):
    # check parameters
    station_code = station.code.upper()
    try:
        ns.check_params(station_code)
        commons.check_env_vars()
    except commons.NSException as e:
        logger.warning(str(e))
        return {"status": "error", "message": str(e)}

    # stop board client if already running
    if request.app.board_client.running:
        request.app.board_client.running = False
        time.sleep(1)

    # start board
    threading.Thread(
        target=request.app.board_client.run,
        args=[
            view_model.ViewModelNS192x32x3ToProtocolBuffers,
            data_updater.DataUpdaterNS,
            {"station_code": station_code},
        ],
    ).start()

    return {"status": "OK"}



from . import data, commons, api


def check_params(station_id: str = None):
    if not any((station_id)):
        return

    stations = data.STATIONS

    if station_id is not None:
        try:
            _ = stations[station_id]
        except Exception as e:
            raise commons.TransilienException(
                f"invalid station code {station_id}"
            ) from e


def stations_by_string(string):
    string = str(string).lower()
    results = {}
    stations = data.STATIONS

    # iterate over stations
    for station_id in stations:
        # match?
        if string in stations[station_id]["nom"].lower():
            results[station_id] = stations[station_id]

    return results


def next_trains(station_id: str):
    # check parameters
    check_params(station_id)

    # get trains departing from station
    departures = api.departures(station_id)

    if departures is None:
        return []

    trains = []

    for train in departures.iter("train"):
        terminus_id = train.find("term").text
        trains.append(
            {
                # extract time from date (formatted as "16/08/2020 20:07")
                "time": train.find("date").text[-5:],
                "mission": train.find("miss").text,
                "terminus_id": terminus_id,
                "terminus": terminus_name(terminus_id),
            }
        )

    return trains


def terminus_name(terminus_id: str) -> str:
    if terminus_id in data.STATIONS:
        return data.STATIONS[terminus_id]["nom"]

    if terminus_id in data.STATIONS_FRANCE:
        return data.STATIONS_FRANCE[terminus_id]["nom"]

    return "???"



from typing import Tuple, List

import departure.board.movement as movement

Pixel = Tuple[int, int, Tuple[int, int, int]]
Pixels = List[Pixel]


def repeated_offsets(
    x_offset: int,
    y_offset: int,
    content_w: int,
    content_h: int,
    display_w: int,
    display_h: int,
    repeat_x: bool = True,
    repeat_y: bool = True,
) -> Tuple[List[int], List[int]]:
    x_offsets = [x_offset]
    y_offsets = [y_offset]

    # repeat right if x_offset + content_w between 0 and display_w
    if repeat_x and (0 < x_offset + content_w < display_w):
        x_offsets.append(x_offset + content_w)

    # repeat down if y_offset + content_h between 0 and display_h
    if repeat_y and (0 < y_offset + content_h < display_h):
        y_offsets.append(y_offset + content_h)

    return x_offsets, y_offsets


def visible_pixels(
    pixels: Pixels, x_offset: int, y_offset: int, display_w: int, display_h: int
) -> Pixels:
    output_pixels = []

    for x, y, c in pixels:
        x_content = x + x_offset
        y_content = y + y_offset
        if (0 <= x_content < display_w) and (0 <= y_content < display_h):
            output_pixels.append((x_content, y_content, c))

    return output_pixels


def visible_pixels_multiple_offsets(
    pixels: Pixels,
    x_offsets: List[int],
    y_offsets: List[int],
    content_w: int,
    content_h: int,
    display_w: int,
    display_h: int,
) -> Pixels:
    output_pixels = []

    for x_offset in x_offsets:
        for y_offset in y_offsets:
            # ignore repetitions that wouldn't be visible
            if (
                (x_offset >= display_w)
                or (x_offset + content_w + display_w < 0)
                or (y_offset >= display_h)
                or (y_offset + content_h + display_h < 0)
            ):
                continue

            output_pixels += visible_pixels(
                pixels, x_offset, y_offset, display_w, display_h
            )

    return output_pixels


class BoardSectionContent:
    def __init__(
        self,
        content_pixels: Pixels,
        content_w: int,
        content_h: int,
        repeat_x: bool = True,
        repeat_y: bool = True,
    ):
        self.content_pixels = content_pixels  # relative to top-left of section

        # virtual bounding box (0, 0, content_w, content_h), used to add space
        # after contents, especially before repeating a message
        self.content_w = content_w
        self.content_h = content_h

        # should content repeat to the right and down?
        self.repeat_x = repeat_x
        self.repeat_y = repeat_y


class BoardSection:
    def __init__(
        self,
        output_x: int,
        output_y: int,
        output_w: int,
        output_h: int,
        content: BoardSectionContent,
        movement: movement.Movement = movement.NoMovement(),
    ):
        # area on the departure board where this section is displayed
        self.output_x = output_x
        self.output_y = output_y
        self.output_w = output_w
        self.output_h = output_h

        # content to be displayed in this section
        self._content = content

        # movement class
        self._movement = movement

        # pixels in absolute coordinates
        self.section_pixels = None

        self._has_changed = True

    def reset(self):
        self._movement.reset()

    def update_content(self, content: BoardSectionContent) -> None:
        self._content = content
        self._has_changed = True

    def update_movement(self, movement: movement.Movement) -> None:
        self._movement = movement
        self._has_changed = True

    def update(self, delta_time: int) -> bool:
        # update movement
        self._has_changed |= self._movement.update(delta_time)

        # wrap offset
        self._wrap_offset()

        return self._has_changed

    def _wrap_offset(self) -> None:
        offset_wrapped = False

        # wrap to between -content_w+1 and 0 if x_offset leq than -content_w
        if self._content.content_w <= 0:
            x_offset = 0
        elif self._movement.x_offset <= -self._content.content_w:
            x_offset = (
                ((self._movement.x_offset - 1) % self._content.content_w)
                - self._content.content_w
                + 1
            )
            offset_wrapped = True
        else:
            x_offset = self._movement.x_offset

        # wrap to between -self._content.content_h+1 and 0 if self._movement.y_offset leq than
        # -self._content.content_h
        if self._content.content_h <= 0:
            y_offset = 0
        elif self._movement.y_offset <= -self._content.content_h:
            y_offset = (
                ((self._movement.y_offset - 1) % self._content.content_h)
                - self._content.content_h
                + 1
            )
            offset_wrapped = True
        else:
            y_offset = self._movement.y_offset

        # update movement offset if it has wrapped
        if offset_wrapped:
            # don't mark as changed as displayed pixels are identical
            self._movement.set_offset(x_offset, y_offset)

    def pixels(self) -> Pixels:
        # return current section pixels if neither content nor movement has
        # changed
        if not self._has_changed:
            return self.section_pixels

        # get repeated offsets
        x_offsets, y_offsets = repeated_offsets(
            self._movement.x_offset,
            self._movement.y_offset,
            self._content.content_w,
            self._content.content_h,
            self.output_w,
            self.output_h,
            self._content.repeat_x,
            self._content.repeat_y,
        )

        # get repeated pixels
        repeated_pixels = visible_pixels_multiple_offsets(
            self._content.content_pixels,
            x_offsets,
            y_offsets,
            self._content.content_w,
            self._content.content_h,
            self.output_w,
            self.output_h,
        )

        # translate pixels to location of board section
        self.section_pixels = []
        for x, y, c in repeated_pixels:
            self.section_pixels.append([x + self.output_x, y + self.output_y, c])

        self._has_changed = False

        return self.section_pixels


class Board:
    def __init__(self, board_w: int, board_h: int, sections: List[BoardSection] = []):
        self.board_w = board_w
        self.board_h = board_h
        self.sections = sections
        self.display_pixels = None
        self._has_changed = True

    def reset(self):
        for section in self.sections:
            section.reset()

    def update(self, delta_time: int):
        for section in self.sections:
            self._has_changed |= section.update(delta_time)
        return self._has_changed

    def pixels(self) -> Pixels:
        if not self._has_changed:
            return self.display_pixels

        self.display_pixels = []
        for section in self.sections:
            self.display_pixels += section.pixels()

        self._has_changed = False

        return self.display_pixels



import time
import threading
import logging

import departure.board.data_updater as data_updater
from departure.commons.log import log_function_stdout_to_debug
from . import view_model, ns, ui, commons

logger = logging.getLogger(__name__)


class DataUpdaterNS(data_updater.DataUpdater):
    def __init__(
        self,
        target_view_model: view_model.ViewModelNS,
        data_refresh_delay_in_s: int,
        end_event: threading.Event,
        station_code: str,
    ):
        super().__init__(target_view_model, data_refresh_delay_in_s, end_event)
        self.station_code = station_code
        self.departures = None

    def update(self):
        update_start_time = time.monotonic()
        logger.info("data refresh started at %s", update_start_time)

        try:
            departures = ns.departures_with_schedule(self.station_code)
        except commons.NSException as e:
            logger.warning("%s - station code %s", str(e), self.station_code)
            return

        # debugging: capture list of services to string, then log
        log_function_stdout_to_debug(
            logger, ui.list_departures_with_schedule, departures, self.station_code
        )

        # update board only if there were changes
        if departures == self.departures:
            logger.info("no change - no update sent to board")
        else:
            self.view_model.update(departures, self.station_code)
            self.departures = departures

        logger.info("data refresh duration %s", time.monotonic() - update_start_time)



import os


class TransilienException(Exception):
    pass


def check_env_vars():
    missing_vars = []
    for transilien_env_var in ["TRANSILIEN_USER", "TRANSILIEN_PASSWORD"]:
        if transilien_env_var not in os.environ:
            missing_vars.append(transilien_env_var)

    if missing_vars:
        raise TransilienException(f"missing env vars {', '.join(missing_vars)}")



import pathlib
import json


DATA_DIRNAME = "data"
IDF_DATA_FILENAME = "sncf-gares-et-arrets-transilien-ile-de-france.json"
FRANCE_DATA_FILENAME = "referentiel-gares-voyageurs.json"


def _stations(data_dirname, idf_data_filename):
    data_path = pathlib.Path(__file__).parents[0] / data_dirname / idf_data_filename

    # read from stations data file
    with open(data_path) as data_file:
        raw_stations = json.load(data_file)

    stations = {}

    for raw_station in raw_stations:
        station_id = raw_station["fields"]["code_uic"]
        station_data = {
            "nom": raw_station["fields"]["nom_gare"],
            "libelle": raw_station["fields"]["libelle"],
        }
        stations[station_id] = station_data

    return stations


def _stations_france(data_dirname, france_data_filename):
    data_path = pathlib.Path(__file__).parents[0] / data_dirname / france_data_filename

    # read from stations data file
    with open(data_path) as data_file:
        raw_stations_france = json.load(data_file)

    stations_france = {}

    for raw_station in raw_stations_france:
        station_id = str(raw_station["fields"]["pltf_uic_code"])
        station_data = {"nom": raw_station["fields"]["gare_alias_libelle_noncontraint"]}
        stations_france[station_id] = station_data

    return stations_france


STATIONS = _stations(DATA_DIRNAME, IDF_DATA_FILENAME)
STATIONS_FRANCE = _stations_france(DATA_DIRNAME, FRANCE_DATA_FILENAME)



"""
Setup for departure module
"""

from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="departure",
    version="1.0.0",
    author="Sébastien Pujadas",
    author_email="sebastien@pujadas.net",
    description="Get station information and departures from public transport "
        "operators, and control virtual or physical departure boards",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/spujadas/departure-python",
    packages=find_namespace_packages(include=["departure.*"]),
    entry_points={
        "console_scripts": [
            "departure=departure.cli.client:entry_point",
            "departure-server=departure.cli.server:entry_point",
            "departure-web=departure.cli.web:start",
        ],
    },
    install_requires=[
        "requests",
        "protobuf",
        "grpcio",
        "bdflib",
        "click",
        "uvicorn",
        "aiofiles",
        "zeep",
        "fastapi",
        "tabulate",
    ],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import os
from zeep import Client
from zeep import xsd


from . import commons


HEADER = xsd.Element(
    "{http://thalesgroup.com/RTTI/2013-11-28/Token/types}AccessToken",
    xsd.ComplexType(
        [
            xsd.Element(
                "{http://thalesgroup.com/RTTI/2013-11-28/Token/types}" "TokenValue",
                xsd.String(),
            ),
        ]
    ),
)

WSDL = "http://lite.realtime.nationalrail.co.uk/OpenLDBWS/wsdl.aspx?ver=2017-10-01"


def soap_client():
    commons.check_env_vars()

    return {
        "headers": [HEADER(TokenValue=os.environ["LDB_TOKEN"])],
        "client": Client(wsdl=WSDL),
    }


def get_departure_board_with_details(station_id: str, num_rows: int = 10):
    soap_client_dict = soap_client()

    return soap_client_dict["client"].service.GetDepBoardWithDetails(
        crs=station_id, numRows=num_rows, _soapheaders=soap_client_dict["headers"]
    )



class Renderer:
    def initialise(self):
        raise NotImplementedError()

    def render_frame(self, pixels):
        raise NotImplementedError()

    def terminate(self):
        raise NotImplementedError()



import os
import json
import logging

import requests


logger = logging.getLogger(__name__)


class AdminException(Exception):
    pass


def board_server_url() -> str:
    # check for DEPARTURE_BOARD_HTTP_PORT env var
    if "DEPARTURE_BOARD_HTTP_PORT" not in os.environ:
        raise AdminException("missing DEPARTURE_BOARD_HTTP_PORT env var")

    # check for DEPARTURE_BOARD_SERVER env var
    departure_board_host = os.environ.get(
        "DEPARTURE_BOARD_SERVER",
        # default: not 'localhost' on Win: https://stackoverflow.com/q/43096253/2654646
        "127.0.0.1",
    )

    return f"http://{departure_board_host}:{os.environ['DEPARTURE_BOARD_HTTP_PORT']}"


def shutdown_board_server() -> dict:
    try:
        url = board_server_url()
    except AdminException as e:
        return {"status": "error", "message": str(e)}

    if "DEPARTURE_BOARD_RSHUTDOWN_TOKEN" not in os.environ:
        return {
            "status": "error",
            "message": "missing DEPARTURE_BOARD_RSHUTDOWN_TOKEN env var",
        }

    # send shutdown request
    try:
        response = requests.post(
            url, json={"token": os.environ["DEPARTURE_BOARD_RSHUTDOWN_TOKEN"]}
        )
    except Exception as e:  # pylint: disable=broad-except
        logger.warning("Error shutting down board server: %s", str(e))
        return {"status": "error", "message": str(e)}

    # return status
    if response.status_code == 200:
        response_json = json.loads(response.content.decode("utf-8"))
        if response_json["status"] == 0:
            return {"status": "OK"}
        return {"status": "error", "message": "invalid token"}

    logger.warning(
        "Error shutting down board server, HTTP response: %s", response.status_code
    )
    return {"status": "error", "message": response.status_code}


def is_board_server_up() -> dict:
    try:
        url = board_server_url()
    except AdminException as e:
        return {"status": "error", "message": str(e)}

    try:
        response = requests.get(url)
    except Exception as e:  # pylint: disable=broad-except
        return {"status": "OK", "server_up": False}

    # return status
    if response.status_code == 200:
        return {"status": "OK", "server_up": True}

    return {"status": "error", "message": response.status_code}


def shutdown_web_server():
    os.system("sudo shutdown -h now")



import time
import threading
import logging

import departure.board.data_updater as data_updater
from departure.commons.log import log_function_stdout_to_debug
from . import view_model, ratp, ui, commons

logger = logging.getLogger(__name__)


class DataUpdaterRatp(data_updater.DataUpdater):
    def __init__(
        self,
        target_view_model: view_model.ViewModelRatp,
        data_refresh_delay_in_s: int,
        end_event: threading.Event,
        line_id: str,
        line_station_id: str,
        direction: str,
    ):
        super().__init__(target_view_model, data_refresh_delay_in_s, end_event)
        self.line_id = line_id
        self.line_station_id = line_station_id
        self.direction = direction
        self.departures = None

    def update(self):
        update_start_time = time.monotonic()
        logger.info("data refresh started at %s", update_start_time)

        try:
            departures = ratp.next_departures(
                self.line_id, self.line_station_id, self.direction
            )
        except commons.RatpException as e:
            logger.warning(
                "%s - line id %s, line station id %s, direction %s",
                str(e),
                self.line_id,
                self.line_station_id,
                self.direction,
            )
            return

        # debugging: capture list of departures to string, then log
        log_function_stdout_to_debug(logger, ui.list_departures, departures)

        # update board only if there were changes
        if departures == self.departures:
            logger.info("no change - no update sent to board")
        else:
            self.view_model.update(departures)
            self.departures = departures

        logger.info("data refresh duration %s", time.monotonic() - update_start_time)



"""
UI for National Rail
"""

from tabulate import tabulate

import departure.commons.helpers as helpers


def list_stations(stations):
    print(
        tabulate(
            [
                [station[0], station[1]]
                for station in sorted(stations.items(), key=lambda k: k[1])
            ],
            headers=["id", "name"],
        )
    )


def list_services(services):
    for i, service in enumerate(services):
        print(
            f"{helpers.ordinal_en(i + 1)} {service['std']} {service['platform']} "
            f"{service['destination_location_name']} "
            f"(from {service['origin_location_name']}) - {service['etd']}"
        )
        calling_points = [
            f"{cp['location_name']} ({cp['st']} / {cp['et']})"
            for cp in service["calling_points"]
        ]
        if i == 0:
            print(f"  Calling at: {', '.join(calling_points)} ({service['operator']})")



import pathlib
import csv


DATA_DIRNAME = "data"
DATA_FILENAME = "stations-2020-01-nl.csv"


def _stations_by_ns_code(data_dirname, data_filename):
    data_path = pathlib.Path(__file__).parents[0] / data_dirname / data_filename

    # read from stations data file
    with open(data_path, encoding="utf8") as data_file:
        # fields: id, code, uic, name_short, name_medium, name_long, slug, country,
        # type, geo_lat, geo_lng
        data_reader = csv.DictReader(data_file, delimiter=",")

        stations = {}

        for station in data_reader:
            stations[station["code"]] = {
                "uic": station["uic"],
                "name_medium": station["name_medium"],
                "name_long": station["name_long"],
            }

    return stations


def _stations_by_uic(stations_by_ns_code_dict):
    stations = {}
    for ns_code in stations_by_ns_code_dict:
        stations[stations_by_ns_code_dict[ns_code]["uic"]] = {
            "ns_code": ns_code,
            "name_medium": stations_by_ns_code_dict[ns_code]["name_medium"],
            "name_long": stations_by_ns_code_dict[ns_code]["name_long"],
        }

    return stations


STATIONS_BY_NS_CODE = _stations_by_ns_code(DATA_DIRNAME, DATA_FILENAME)
STATIONS_BY_UIC = _stations_by_uic(STATIONS_BY_NS_CODE)



"""
CLI for Transilien
"""

import threading
import time
import logging

import click

from departure.board import board_client

from . import ui, transilien, commons, view_model, data_updater

ONE_DAY_IN_SECONDS = 60 * 60 * 24
logger = logging.getLogger(__name__)


@click.group()
def cli():
    """
    Information and departures for National Rail (UK).

    Note - Your Transilen API user and password must be assigned to the TRANSILIEN_USER
    and TRANSILIEN_PASSWORD environment variables.
    """


@click.command()
@click.argument("query_string")
def search(query_string):
    """Search for stations containing QUERY_STRING."""
    stations = transilien.stations_by_string(query_string)
    ui.list_stations(stations)


@click.command(name="next")
@click.argument("station_id")
def next_trains(station_id):
    """
    Get next trains at station STATION_ID.

    Use search to find the STATION_ID for a station.
    """

    try:
        trains = transilien.next_trains(station_id)
    except commons.TransilienException as e:
        print(f"error: {str(e)}")
        return

    ui.list_trains(trains)


@click.command(name="board")
@click.argument("station_id")
def start_board(station_id):
    """
    Update a departure board with trains at station STATION_ID.

    Use search to find the STATION_ID for a station.
    """

    try:
        transilien.check_params(station_id)
    except commons.TransilienException as e:
        print(f"error: {str(e)}")
        return

    # start board
    board_client_instance = board_client.BoardClient()

    threading.Thread(
        target=board_client_instance.run,
        args=[
            view_model.ViewModelTransilien_192_32_3_Rows_To_ProtocolBuffers,
            data_updater.DataUpdaterTransilien,
            {
                "station_id": station_id,
            },
        ],
    ).start()

    # run until interrupt
    try:
        while True:
            time.sleep(ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        logger.info("received interrupt")
        board_client_instance.running = False


cli.add_command(search)
cli.add_command(next_trains)
cli.add_command(start_board)

if __name__ == "__main__":
    cli()



import logging

import grpc

import departure.board.view_model as view_model
import departure.board.contents as contents
import departure.commons.helpers as helpers
import departure.board.departure_pb2 as departure_pb2
import departure.board.departure_pb2_grpc as departure_pb2_grpc  # for type hinting
import departure.board.protobuf as board_protobuf

logger = logging.getLogger(__name__)


class ViewModelRatp(view_model.ViewModel):  # pylint: disable=abstract-method
    pass


class ViewModelRatp_192_32(ViewModelRatp):  # pylint: disable=abstract-method
    def __init__(self):
        self.board_text = contents.BoardText(
            font=contents.read_bdf_font("fonts/6x10_condensed.bdf"),
            colour_map={
                "orange": (255, 204, 0),
                "red": (255, 0, 0),
                "green": (0, 255, 0),
                "white": (255, 255, 255),
            },
            tabs=[[0, "l"], [18, "l"], [42, "l"], [191, "r"]],
        )

    def text_array_service_overview(self, mission, pos: int):
        if mission["code"] is None:  # metro, bus
            return [
                [helpers.ordinal_fr(pos), "orange"],
                # [mission['message'] + ' - ' + mission['destinationName'], 'orange'],
                [mission["destinationName"], "orange"],
                ["", "orange"],
                [mission["message"], "white"],
            ]
        else:  # RER
            return [
                [helpers.ordinal_fr(pos), "orange"],
                [mission["code"], "white"],
                [mission["message"] + " - " + mission["destinationName"], "orange"]
                # [mission['destinationName'], 'orange']
            ]

    def pixels_mission_overview(self, mission, pos: str):
        return self.board_text.colour_text_tabbed_row_pixels(
            self.text_array_service_overview(mission, pos)
        )

    def next_content(self, departures):
        # empty board if called with None
        if departures is None:
            return [], (0, 0), [], (0, 0), [], (0, 0)

        # top section
        if len(departures["missions"]) == 0:  # "Pas de service" if no service
            (
                top_section_content_pixels,
                top_section_content_pixels_size,
            ) = self.board_text.colour_text_pixels([["Pas de service", "orange"]])
        else:  # otherwise: 1st mission
            (
                top_section_content_pixels,
                top_section_content_pixels_size,
            ) = self.pixels_mission_overview(departures["missions"][0], 1)

        # middle section
        if len(departures["missions"]) < 2:  # nothing if fewer than 2 trains
            middle_section_content_pixels = []
            middle_section_content_pixels_size = (0, 0)
        else:  # otherwise: 2nd mission
            (
                middle_section_content_pixels,
                middle_section_content_pixels_size,
            ) = self.pixels_mission_overview(departures["missions"][1], 2)

        # bottom section
        if departures["perturbations"]:  # disruptions if any
            (
                bottom_section_content_pixels,
                bottom_section_content_pixels_size,
            ) = self.board_text.colour_text_pixels(
                [[departures["perturbations"], "red"]]
            )
            # adjust space before repeating
            bottom_section_content_pixels_size = (
                bottom_section_content_pixels_size[0] + 192,
                bottom_section_content_pixels_size[1],
            )

        elif len(departures["missions"]) < 3:  # nothing if fewer than 3 trains
            bottom_section_content_pixels = []
            bottom_section_content_pixels_size = (0, 0)

        elif len(departures["missions"]) == 3:  # 3rd train (if only 3 trains)
            (
                bottom_section_content_pixels,
                bottom_section_content_pixels_size,
            ) = self.pixels_mission_overview(departures["missions"][2], 3)

        else:  # 3rd train onwards (if more than 3 trains)
            (
                bottom_section_content_pixels,
                bottom_section_content_pixels_size,
            ) = self.board_text.colour_text_tabbed_rows_pixels(
                [
                    self.text_array_service_overview(mission, i + 3)
                    for i, mission in enumerate(departures["missions"][2:])
                ],
                row_height=11,
            )

        return (
            top_section_content_pixels,
            top_section_content_pixels_size,
            middle_section_content_pixels,
            middle_section_content_pixels_size,
            bottom_section_content_pixels,
            bottom_section_content_pixels_size,
        )


class ViewModelRatp_192_32_3_Rows_To_ProtocolBuffers(ViewModelRatp_192_32):
    def __init__(self, board_manager_stub: departure_pb2_grpc.BoardManagerStub):
        super().__init__()
        self.board_manager_stub = board_manager_stub

        self.next_mission_tracker = [
            departure_pb2.Movement(
                static_content=departure_pb2.StaticContent(total_duration=2000)
            ),
            departure_pb2.Movement(
                scrolling_content=departure_pb2.ScrollingContent(
                    step_duration=100,
                    total_steps=11,
                    delta_x_per_step=0,
                    delta_y_per_step=-1,
                )
            ),
        ]

        self.horizontal_scroll = [
            departure_pb2.Movement(
                scrolling_content=departure_pb2.ScrollingContent(
                    step_duration=50,
                    total_steps=0,
                    delta_x_per_step=-1,
                    delta_y_per_step=0,
                )
            )
        ]

        self.no_movement = [
            departure_pb2.Movement(no_movement=departure_pb2.NoMovement())
        ]

    def update(self, departures):  # pylint: disable=arguments-differ
        requests_data = []

        (
            top_section_content_pixels,
            top_section_content_pixels_size,
            middle_section_content_pixels,
            middle_section_content_pixels_size,
            bottom_section_content_pixels,
            bottom_section_content_pixels_size,
        ) = self.next_content(departures)

        # top section: 1st train
        requests_data.append(
            departure_pb2.BoardSectionUpdateRequest(
                section_index=0,
                content=departure_pb2.BoardSectionContent(
                    pixels=board_protobuf.serialise_pixels(top_section_content_pixels),
                    content_w=top_section_content_pixels_size[0],
                    content_h=top_section_content_pixels_size[1],
                    repeat_x=False,
                    repeat_y=False,
                ),
                movement=self.no_movement,
                continue_movement=False,
            )
        )

        # middle section: 2nd train
        requests_data.append(
            departure_pb2.BoardSectionUpdateRequest(
                section_index=1,
                content=departure_pb2.BoardSectionContent(
                    pixels=board_protobuf.serialise_pixels(
                        middle_section_content_pixels
                    ),
                    content_w=middle_section_content_pixels_size[0],
                    content_h=middle_section_content_pixels_size[1],
                    repeat_x=False,
                    repeat_y=False,
                ),
                movement=self.no_movement,
                continue_movement=False,
            )
        )

        # bottom section: 3rd train onwards or disruptions
        repeat_x = False
        repeat_y = False

        if departures is None:
            next_movement = self.no_movement
        elif departures["perturbations"]:  # h-scroll if disruptions
            next_movement = self.horizontal_scroll
            repeat_x = True
        elif len(departures["missions"]) <= 3:  # no v-scroll if up to 3 trains
            next_movement = self.no_movement
        else:  # v-scroll if more than 3 trains
            next_movement = self.next_mission_tracker
            repeat_y = True

        requests_data.append(
            departure_pb2.BoardSectionUpdateRequest(
                section_index=2,
                content=departure_pb2.BoardSectionContent(
                    pixels=board_protobuf.serialise_pixels(
                        bottom_section_content_pixels
                    ),
                    content_w=bottom_section_content_pixels_size[0],
                    content_h=bottom_section_content_pixels_size[1],
                    repeat_x=repeat_x,
                    repeat_y=repeat_y,
                ),
                movement=next_movement,
                continue_movement=True,
            )
        )

        # send gRPC message
        try:
            self.board_manager_stub.BoardSectionsUpdate(
                departure_pb2.BoardSectionsUpdateRequest(requests=requests_data)
            )
        # pylint: disable=protected-access
        except (grpc._channel._Rendezvous, grpc._channel._InactiveRpcError):
            logger.warning("connection to board failed")



import importlib
import pkgutil

import click

from departure.commons.log import init_logging

# initialise logging
init_logging()


# CLI entry point
@click.group()
def entry_point():
    """
    Starts a departure board display server using the back end specified by COMMAND.
    Back ends are available as additional packages, see online documentation for full
    list.
    """


# find modules in namespace
def iter_namespace(ns_pkg):
    # code from https://packaging.python.org/guides/creating-and-discovering-plugins/
    # Specifying the second argument (prefix) to iter_modules makes the
    # returned name an absolute name instead of a relative one. This allows
    # import_module to work without having to do additional modification to
    # the name.
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


# import renderers if any
try:
    # ignore pylint errors if no renderers available => will be caught at runtime
    # pylint: disable=no-name-in-module,import-error
    import departure.renderer

    # find available renderers (departure.renderer.*) and add their command to main CLI
    # pylint: disable=no-member
    for finder, name, ispkg in iter_namespace(departure.renderer):
        # get .cli submodule
        submodule = importlib.import_module(".cli", name)

        # get run() function - bound to @click.command
        command_function = getattr(submodule, "run")

        # get command name for CLI
        command_name = getattr(submodule, "COMMAND")

        # add command to CLI
        entry_point.add_command(command_function, name=command_name)
except ModuleNotFoundError:
    # run anyway if no renderers found
    pass


if __name__ == "__main__":
    entry_point()
